#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import SCHEMA_LOCN_TEMPL 
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlreports.webReportUtils import SITE_DIR_TEMPL, TOP_MD_TEMPL, keyStats, roundFloat
from fmqlreports.webReportUtils import muPlotRef, makePlots, vistasOfVISNByOne

from fileDirectory import CommonFileDirectory

"""
To get to a "Core" VistA and "Exclusive" VistAs for all VISN sites. 

TODO: 
--------------------- SPECIFIC CONTENT ---------------------------
- burrow more into Exclusives (particularly in Puget) <------ BIG ONE
- 'some' files -- do SPO and POR have populated overlaps as only two to jump out for common in last graph
- verify empties can be ignored
--------------------- GENERIC/CACHE RELATED ----------------------
- consider singleton (at most one where appears ie/ can be 0 too but at most one) descriptions as these are configs?
- bands of file sizes -- # of files in bands of size ... [part of Caching]
- mult #'s ie/ # contained ... indication that doc, not rec?
- old count of pter vs ... ie/ get into quantities of props
- CACHED sizes (est vs real) [message: much variety but not big size]
"""

def webReportFileDirectory(stationNoMap, imageDir="Images"):
    
    cfd = CommonFileDirectory(stationNoMap)
    allFLs = cfd.fileIds()
    ttlRecords = cfd.totalRecords()

    mu = TOP_MD_TEMPL.format("VISN 20 File Directory")
        
    mu += """# VISN 20 File Directory
    
_FileMan_ is VistA's proprietary database. It holds not only clinical data but also nearly all the configurations, audit logs and financial information of the system. Having all this data in one location is a boon for VistA system analysis and the centrality of FileMan effectively turns VistA analysis into FileMan content analysis. 

In _FileMan_, data is gathered into \"files\" which are structured by \"data dictionaries\" or schemas. 
    
The following reports on a VISN 20 File Directory built from the data dictionaries of 7 VistAs: Puget Sound [663], Portland [648], Boise [531], Spokane [668], Roseburg [653], White City [692] and Walla Walla [687]. Together the systems hold <span class='yellowIt'>{:,}</span> distinct files containing <span class='yellowIt'>{:,}</span> records.

Note: Anchorage is the missing VISN 20 system. Its information will be added once the system is made available.
    
""".format(
        len(allFLs),
        ttlRecords
    )

    plotData = {}
    
    # ######################## Summary Criteria ############################
    # 
    # ... exclusives, singletons, all's, empties ... 
    #
    
    commonCount = cfd.commonCounts()            
    allCnt = max(commonCount)
    plotData["fileTypes"] = {
            "title": "All File Types",
            "plotName": "visn20AllFileTypes",
            "plotMethod": "plotIsHasBH",
            "specs": {
                "__entityName": "Files",
                "__total": len(allFLs),
                "all": commonCount[allCnt],
                "some": sum(commonCount[i] for i in range(2, allCnt)),
                "exclusive": commonCount[1] # rem 1-# of VistAs
            }
    }
    blurb = "Files can be grouped into those in every VistA (\"all\"), those in more than one but not all (\"some\") and those only in one (\"exclusive\"). The files in every VistA represent the __core of a VA VistA__, that portion of FileMan common to production systems ..."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["fileTypes"], imageDir)
    )

    emptiesInAll = set(cfd.filesOfCount(count=0))
    singletons = set(cfd.filesOfCount(count=1))
    plotData["fileTypesEmptySingleton"] = {
            "title": "File Types in Use",
            "plotName": "visn20FilesTypesInUse",
            "plotMethod": "plotIsHasBH",
            "specs": {
                "__entityName": "Files",
                "__total": len(allFLs),
                "empty in all": len(emptiesInAll),
                "singleton in all": len(singletons)
            }
    }
    # TODO: could make singleton == 0 or at most 1?
    # TODO: retired => last record from 5 plus years ago?
    blurb = "Files can also be grouped on whether they are \"empty\" or have only one record (\"singletons\") whereever they appear. The _empties_ usually represent retired parts of VistA and can be put aside ..."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["fileTypesEmptySingleton"], imageDir)
    )
    
    # ####################### Overall Proportions less Empty ###################
    
    commonCountNoEmpties = cfd.commonCounts(excludeFLs=emptiesInAll)
    alls = set(cfd.commonFiles())
    exclFiles = cfd.commonFiles(count=1) # only appear in one
    allFlsByGroup = {}
    for fl in allFLs:
            if fl in emptiesInAll:
                continue
            if fl in exclFiles:
                allFlsByGroup[fl] = "exclusive"
                continue
            if fl in alls:
                allFlsByGroup[fl] = "all"
                continue
            allFlsByGroup[fl] = "some"
    totalsByGroup = cfd.totalRecords(groupOfFile=allFlsByGroup)
    
    grpColumns = ["all", "some", "exclusive"]
    plotData["filesNRecordsByFileCategory"] = {
            "title": "File and Record Counts by File Category",
            "plotName": "visn20FileNRecordCountsByFileType",
            "plotMethod": "plotCategoryBH",
            "rows": ["files", "records"],
            "columns": grpColumns,
            "data": [
                (
                    commonCountNoEmpties[allCnt],
                    sum(commonCountNoEmpties[i] for i in range(2, allCnt)),
                    commonCountNoEmpties[1]
                ),
                (
                    totalsByGroup["all"],
                    totalsByGroup["some"],
                    totalsByGroup["exclusive"]
                )
            ]
    }
    blurb = "Putting aside the empty files puts focus on the active parts of the VISN 20 VistAs. See the contrast between common or core files and the proportion of records in those files - while only <span class='yellowIt'>{}</span> files are common to all systems, those common files contain <span class='yellowIt'>{}</span> of the records in these systems. In other words, the core of a VistA FileMan holds most of the records ...".format(
        reportAbsAndPercent(
            plotData["filesNRecordsByFileCategory"]["data"][0][0],
            sum(commonCountNoEmpties[i] for i in range(0, allCnt+1))
        ),
        reportPercent(
            plotData["filesNRecordsByFileCategory"]["data"][1][0],
            sum(totalsByGroup[c] for c in totalsByGroup)
        )
    )
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["filesNRecordsByFileCategory"], imageDir)
    )
    
    # ####################### By VistA (4) Common Details ###################

    commonByVistA = defaultdict(dict) # not picturing           
    commonNoEmptiesByVistA = defaultdict(dict)
    totalsByGroupByVistA = {}
    for vistaInfo in cfd.vistas():
            stationNo = vistaInfo["stationNumber"]
            stationMN = stationNoMap[stationNo]
            commonCount = cfd.commonCounts(stationNo)
            commonByVistA[stationMN] = {
                "all": commonCount[allCnt],
                "some": sum(commonCount[i] for i in range(2, allCnt)),
                "exclusive": commonCount[1]
            }
            commonCountNoEmpties = cfd.commonCounts(stationNo, emptiesInAll)
            commonNoEmptiesByVistA[stationMN] = {
                "all": commonCountNoEmpties[allCnt],
                "some": sum(commonCountNoEmpties[i] for i in range(2, allCnt)),
                "exclusive": commonCountNoEmpties[1]
            }
            mnTotal = cfd.totalRecords(stationNo, groupOfFile=allFlsByGroup)
            totalsByGroupByVistA[stationMN] = dict((grp, mnTotal.get(grp, 0)) for grp in grpColumns)

    # may have "flip perc" as option
    plotData["commonNoEmptiesByVistA"] = {
            "title": "VISN 20 VistAs - File Types Compared (No Empties)",
            "plotName": "visn20SideBySideCommonCountsNoEmpties",
            "plotMethod": "plotCategoryBSV",
            "specs": commonNoEmptiesByVistA
    }
    blurb = "Put side by side, you see that systems big and small have very similar numbers of files but that larger systems like Portland (\"POR\") have proportionately more exclusive files ..."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["commonNoEmptiesByVistA"], imageDir)
    )
    
    plotData["totalRecordsByCommonByVistA"] = {
            "title": "VISN 20 VistAs - Records",
            "plotName": "visn20SideBySideByCommonRecordCounts",
            "plotMethod": "plotCategoryBSV",
            "specs": totalsByGroupByVistA
    }
    blurb = "Record numbers, side by side tell a different story. The biggest system is Puget - not as file count indicates, Portland - and Puget is responsible for most of the populated exclusive files. In other words, though all systems have some exclusive files, they are of most (only of?) importance in Puget ..."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["totalRecordsByCommonByVistA"], imageDir)
    )
    
    # ###################### Save Plot Data and .md #####################
        
    plotDir = f'{VISTA_DATA_BASE_DIR}Common/TmpWorking/'
    print(f'Serializing vizData to {plotDir}')
    json.dump(plotData, open(f'{plotDir}plotDataFiles.json', "w"), indent=4)
    
    siteDir = SITE_DIR_TEMPL.format("Common")
    print(f'Serializing Report to {siteDir}')
    open(f'{siteDir}visn20FileDirectory.md', "w").write(mu) 

# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 4)
    
    if len(sys.argv) > 1:
        if sys.argv[1] == "PLOT":
            makePlots("Common", "Files")
            return
        raise Exception("Only argument allowed is 'PLOT'")

    webReportFileDirectory(vistasOfVISNByOne("20"))    
                 
if __name__ == "__main__":
    main()


#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue
from ..webReportUtils import SITE_DIR_TEMPL, loadSelectTypes, reduce200, mdFilesOfDomain

"""
Prosthetics ... follow Dental, Health Factors

Requires:
- SELECT TYPEs for File Types

TO SEE:
- home ox Rx in multiple of Pros Patient

"""
def webReportProsthetics(stationNo):

    allThere, details = checkDataPresent(stationNo, [
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    mu = """---
layout: default
title: {} Prosthetics
---

## Prosthetics

""".format(stationNo) 
    
    usedFiles = ["PROSTHETICS PATIENT"]
    mu += mdFilesOfDomain(
        stationNo, 
        "Prosthetics", 
        660, 
        670,
        usedFiles
    )
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "prosthetics.md", "w").write(mu)
        
# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 4)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise SystemExit("Expect User Site to already exist with its basic contents")
    
    webReportProsthetics(stationNo)
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime
import platform
import statistics

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import FilteredResultIterator 
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlreports.webReportUtils import vistasOfVISNByOne

from fileDirectory import CommonFileDirectory

"""
TODO:
- add in first install, # installs, last install per VistA of build (and can then graph installs vs ...) + simple cnt of any builds NOT installed => exclude from totals
- entries with missing or inconsistent dates ... report on it (see below warnings) [210 warnings from 13197]

TODO STRUCT/INSTALL: common base directory. Moving towards that with 'resource' and 'entry'. Just need to establish the base that WORKS FOR BUILDS AND FILES (as both 
unique in space)
REM: goal is that basic report drops out common.
REM: idea of set of resources with entries in each VistA (one per VistA or could have
> one in case of users?)
"""
class CommonBuildDirectory(): 

    ENTRY_IDX_SNO = 0

    ENTRY_IDX_ID = 1 # IEN
    ENTRY_IDX_NAME = 2 # assuming unique
    
    ENTRY_IDX_DATE_DISTRIBUTED = 3
    ENTRY_IDX_PACKAGE = 4
    ENTRY_IDX_TYPE = 5
    ENTRY_IDX_DESCRIPTION = 6
    ENTRY_IDX_ISNATIONAL = 7

    ENTRY_IDX_FILES = 8
    
    # Portable
    ENTRY_IDX_RESOURCE_ID = 2 # CommonBuildDirectory.ENTRY_IDX_NAME
    
    """
    # BUILD/INSTALL on file -- see if appropriate
    ENTRY_IDX_BUILD_FIRST_DT = 6
    ENTRY_IDX_BUILD_LAST_DT = 7
    ENTRY_IDX_BUILD_COUNT = 8
    ENTRY_IDX_INSTALL_FIRST_DT = 9
    ENTRY_IDX_INSTALL_LAST_DT = 10
    ENTRY_IDX_INSTALL_COUNT = 11
    """

    def __init__(self, vistaSNOs, dirPrefixLetter="B"):
        self.__vistaSNOs = vistaSNOs
        self.__dirFile = f'{VISTA_DATA_BASE_DIR}Common/Directory/{dirPrefixLetter}DIR.json'
        try:
            print("Attempting to load preexisting Directory")
            self.__dir = json.load(open(self.__dirFile))
        except:
            print("Starting Directory from Scratch")
            self.__dir = {"vistas": [], "entries": {}, "resources": []}
        else:
            print(f'Loaded preexisting directory with {len(self.__dir["resources"]):,} builds')
        self.__changeUpdate()
        
    def vistas(self):
        return self.__dir["vistas"]
        
    def resourceIds(self, stationNo): # to be portable
        return self.buildIds(stationNo)
        
    # To see, per station, add stationNo
    def total(self, stationNo=""):
        if stationNo == "":
            return len(self.__dir["resources"])
        return sum(1 for res in self.__dir["resources"] if sum(1 for gien in res if gien.split(":")[0] == stationNo))
        
    def entryTotal(self, stationNo=""):
        if stationNo == "":
            return len(self.__dir["entries"])
        return sum(1 for gien in self.__dir["entries"] if gien.split(":")[0] == stationNo)
                
    def lookupEntryByGIEN(self, gien):
        return self.__dir["entries"].get(gien, None)
        
    def commonCounts(self, stationNo="", excludeResources=set()):
        crossVistATypesPresenceCount = Counter()
        for res in self.__dir["resources"]:
            resName = self.__dir["entries"][res[0]][CommonBuildDirectory.ENTRY_IDX_RESOURCE_ID]
            crossVistATypesPresenceCount[resName] = len(res)
        ress = self.resourceIds(stationNo) # all if stationNo == ""
        commonCount = Counter()
        for resId in ress:
            if resId in excludeResources:
                continue
            commonCount[crossVistATypesPresenceCount[resId]] += 1                    
        return commonCount
                
    # Fan out the build - good for reports
    def lookupBuildEntriesByGIEN(self, gien):
        gienBuild = None
        for build in self.__dir["resources"]:
            if gien in build:
                gienBuild = build
                break
        if not gienBuild:
            return None
        entries = []
        for gien in gienBuild:
            entry = self.__dir["entries"][gien]
            entries.append(entry)
        return {"gien": gien, "entries": entries}
               
    """
    Plan to support report with: top like file rep, time like user
    - hbar of total, excl, common, some
    - hbar of w/files (and can add w'routines later)
    - TODO: number of files covered by build system [could you rebuild a VistA]
    - common counts per vista, abs and perc ... excl, all, some
    - 
    """         
    def report(self, fileDirectory=None):        
        
        mu = "# Build Directory Synopsis\n\n"
        mu += f'VistAs: {len(self.__dir["vistas"]):,}' + "\n"
        
        # Graphics - top: topTotal - isHasBH
        ress = self.__dir["resources"]
        mu += f'Builds: {len(self.__dir["resources"]):,}' + "\n"
        cores = set(self.commonBuilds())
        mu += "Core (in all): {}\n".format(
            reportAbsAndPercent(len(cores), len(ress))
        )
        commonCount = self.commonCounts()
        mu += f'Common Builds: {"/".join([reportAbsAndPercent(commonCount[cnt], len(ress)) for cnt in sorted(commonCount) if cnt > 1])}' + "\n"
        exclBuilds = self.commonBuilds(count=1) # if they appear, are one
        mu += "Exclusive (in 1): {}\n".format(
            reportAbsAndPercent(len(exclBuilds), len(ress)),
        )
                
        # 1. First Bar up top breakdown [+ per VistA, will do in Graph]
        allBldsByGroup = {}
        for resId in exclBuilds:
            allBldsByGroup[resId] = "EXCLUSIVE"
        for resId in cores:
            allBldsByGroup[resId] = "ALL"
        for resId in self.buildIds():
            if resId in allBldsByGroup:
                continue
            allBldsByGroup[resId] = "SOME"
        if "SOME" in allBldsByGroup:
            mu += f"In Some (not just 1, not all): {reportAbsAndPercent(len(allBldsByGroup['SOME']), len(ress))}\n"
        
        # TODO: Can expand to have routines etc
        # ... 2. Second Bar up top
        mu += f'Have files: {reportAbsAndPercent(len(self.buildsWithFiles()), len(ress))}\n'
        entryDateByBuild = self.buildEntryDates()
        mu += f'Have Entry Dates: {reportAbsAndPercent(len(entryDateByBuild), len(ress))}\n'
        mu += f'Have packages: {reportAbsAndPercent(len(self.buildsWithPackage()), len(ress))}\n'        
        
        # Timing - will have builds by year (ala last bar in user or patient)
        countByYr = Counter()
        for resName in entryDateByBuild:
            yr = entryDateByBuild[resName].split("-")[0]
            countByYr[yr] += 1
        mu += "Builds by Year (of builds with entry date) ...\n"
        for yr in sorted(countByYr, key=lambda x: int(x)):
            mu += f"\t{yr}: {countByYr[yr]}\n"
        mu += "\n"
        
        # 3. TODO: 
        # expand out and see % of known files covered or addressed in builds
        mu += f'Files: {len(self.buildFiles())}\n'
        if fileDirectory:
            mu += f'Files (% of all known): {reportAbsAndPercent(len(self.buildFiles()), len(fileDirectory.fileIds()))}\n'
        else:
            mu += f'Files: {len(self.buildFiles())}\n'            
        mu += f'Packages: {len(self.buildPackages())}\n'
            
        mu += f'Definitions across VistAs: {len(self.__dir["entries"]):,}' + "\n"
        mu += "\n"
        for vistaInfo in self.__dir["vistas"]:
        
            stationNo = vistaInfo["stationNumber"]
            mu += f'## VistA {stationNo}\n' + "\n"
            
            ress = self.buildIds(stationNo)
            mu += f'Builds: {len(ress):,}' + "\n"
            commonCount = self.commonCounts(stationNo)
            mu += f'Common Builds: {"/".join([reportAbsAndPercent(commonCount[cnt], len(ress)) for cnt in sorted(commonCount) if cnt > 1])}' + "\n"
            exclBuilds = set(self.commonBuilds(stationNo, count=1))
            mu += "Exclusive (in 1): {}\n".format(
                reportAbsAndPercent(len(exclBuilds), len(ress)),
            )
            mu += "\n" 
                
        return mu
        
    # ----------------- Build Ids and Presence across VistAs -------------------
            
    def buildIds(self, stationNo=""):
        """
        All builds known or just per VistA - rem id is name
        
        For Plot: enables a isHasBH - total builds and then can give % with different 
        attributes using other methods below (emptyBuildIds)
        """
        resNames = []
        if stationNo:
            for entry in self.__entriesByVistA(stationNo):
                resNames.append(entry[CommonBuildDirectory.ENTRY_IDX_NAME])
        else:    
            for res in self.__dir["resources"]:
                # assuming all in res have same id == name
                entry = self.__dir["entries"][res[0]]
                resNames.append(entry[CommonBuildDirectory.ENTRY_IDX_NAME])
        return sorted(resNames)
        
    def commonBuilds(self, stationNo="", count=0, mustHaveFiles=False): 
        """
        # 0 == ALL
            ie/ commonBuilds(stationNo) <=> allBuilds([stationNo])
        # 1 == Exclusive (ie/ not common at all)
            ie/ commonBuilds(stationNo, 1) <=> exclusiveBuilds([stationNo])
        where stationNo being set limits the list to builds in a particular vista
        
        # > 1 BUT < len(vistas) => mix ie/ not exclusive and not all
        """
        inCount = []
        count = count or len(self.__dir["vistas"])
        for res in self.__dir["resources"]:
            if len(res) != count: # relies on builds only once in any vista
                continue
            if stationNo and sum(1 for b in res if b.split(":")[0] == stationNo) == 0:
                continue
            resEntry = self.__dir["entries"][res[0]]
            if mustHaveFiles and len(resEntry[CommonBuildDirectory.ENTRY_IDX_FILES]) == 0:
                continue
            resName = resEntry[CommonBuildDirectory.ENTRY_IDX_NAME]
            inCount.append(resName)
        return inCount
        
    def buildFiles(self):
        allFLs = set()
        for gien in self.__dir["entries"]:
            bentry = self.__dir["entries"][gien]
            fls = bentry[CommonBuildDirectory.ENTRY_IDX_FILES]
            if len(fls) == 0:
                continue
            allFLs |= set(fls)
        return sorted(list(allFLs))

    def buildsWithFiles(self):
        wFiles = []
        for res in self.__dir["resources"]:
            bentry = self.__dir["entries"][res[0]]
            if len(bentry[CommonBuildDirectory.ENTRY_IDX_FILES]):
                resName = bentry[CommonBuildDirectory.ENTRY_IDX_NAME]
                wFiles.append(resName)
        return wFiles
        
    def buildPackages(self):
        allPKGs = set()
        for gien in self.__dir["entries"]:
            bentry = self.__dir["entries"][gien]
            pkg = bentry[CommonBuildDirectory.ENTRY_IDX_PACKAGE]
            if pkg == "":
                continue
            allPKGs.add(pkg)
        return sorted(list(allPKGs))
        
    def buildsWithPackage(self):
        wPackages = []
        for res in self.__dir["resources"]:
            bentry = self.__dir["entries"][res[0]]
            if bentry[CommonBuildDirectory.ENTRY_IDX_PACKAGE]:
                resName = bentry[CommonBuildDirectory.ENTRY_IDX_NAME]
                wPackages.append(resName)
        return wPackages
        
    """
    Dates by Build Name
    
    Expecting consistency of entry dates - ie/ one per build
    
    Note: will go with build install per VistA dates
    """
    def buildEntryDates(self):
        entryDateOfBuild = {}
        warning = defaultdict(dict)
        for res in self.__dir["resources"]:
            resDT = ""
            for gien in res:
                entry = self.__dir["entries"][gien]
                resName = entry[CommonBuildDirectory.ENTRY_IDX_NAME]
                dt = entry[CommonBuildDirectory.ENTRY_IDX_DATE_DISTRIBUTED]
                if dt:
                    if resDT:
                        if resDT != dt:
                            warning["DATE_DISAGREE"][resName] = f'{resDT} vs {dt}'
                            if dt < resDT:
                                resDT = dt
                    else:
                        resDT = dt
                        entryDateOfBuild[resName] = dt    
                elif resDT:
                    warning["DATE_THERE_AND_NOT"][resName] = entry[CommonBuildDirectory.ENTRY_IDX_SNO]
        if len(warning):
            print(f'** Warning: for date extractions out of {len(self.__dir["resources"])}, {sum(len(warning[wt]) for wt in warning)} warnings happen')
        return entryDateOfBuild

    # -------------------------- Utilities etc --------------------------
        
    def __entriesByVistA(self, stationNo):
        return [self.__dir["entries"][gid] for gid in self.__dir["entries"] if re.match(stationNo, gid)]
        
    # TODO: timing stuff is ever get to 9_4, 9_6 - see old Archives
             
    # ########################## CHANGE / UPDATE ######################
                        
    def __changeUpdate(self):
        changed = self.__fillFromAvailable()
        if changed:
            print("Changed - so rebuilding builds")
            self.__rebuildBuilds()
            self.__flush()    
                             
    def __fillFromAvailable(self):
    
        def loadStationBuilds(stationNo):
    
            dataLocn = "{}{}/Data/".format(VISTA_DATA_BASE_DIR, stationNo)   
            resourceIter = FilteredResultIterator(dataLocn, "9_6")
            resourceEntry = []
            fileAssertionMismatchCounter = Counter()
            wFiles = 0
            for resource in resourceIter:
                """
                Add:
                - build components
                - multiple build
                - required build
                # DO: build_components - 
                """
                flIds = []
                if "file" in resource:
                    fls = sorted(list(set(mult["ien"] for mult in resource["file"])), key=lambda x: float(x))         
                    flIds = sorted(list(set(mult["file"]["id"].split("-")[1] for mult in resource["file"] if "file" in mult)), key=lambda x: float(x))
                    wFiles += 1
                    if len(set(flIds) - set(fls)):
                        raise Exception("If iens and fl pointers differ then expect iens to be the superset")
                    if flIds != fls:
                        fileAssertionMismatchCounter[len(fls) - len(flIds)] += 1
                rresource = [
                    stationNo,
                    resource["_id"].split("-")[1],
                    resource["name"],
                    resource["date_distributed"]["value"] if "date_distributed" in resource else "",
                    resource["package_file_link"]["label"] if "package_file_link" in resource else "",
                    resource["type_2"].split(":")[0] if "type_2" in resource else "",
                    resource.get("description_of_enhancements", ""),
                    resource.get("track_package_nationally", ""),
                    flIds # taking subset of mults - these ones exist in 1/DD
                ]
                self.__dir["entries"][f'{stationNo}:{rresource[1]}'] = rresource
            if wFiles:
                print(f"** Warning {stationNo}: 'file' pointer missing from {fileAssertionMismatchCounter} of the {wFiles} entries with files") 
            
        changed = False
        for stationNo in self.__vistaSNOs:
            if sum(1 for vistaInfo in self.__dir["vistas"] if vistaInfo["stationNumber"] == stationNo):
                print("\t{} already in Directory - skipping".format(stationNo))
                continue
            try:
                loadStationBuilds(stationNo)
            except Exception as e:
                print("\t{} can't be loaded as no 9_6 ...".format(stationNo))
                print(e)
                continue
            changed = True
            self.__dir["vistas"].append(
                {
                    "stationNumber": stationNo, 
                    "dateAdded": str(datetime.now()).split(" ")[0]
                }
            )
        return changed
                
    def __rebuildBuilds(self): # Just doing top builds for now
        del self.__dir["resources"]
        buildEntryGIDsByName = defaultdict(list)
        for gid in self.__dir["entries"]:
            entry = self.__dir["entries"][gid]
            buildEntryGIDsByName[entry[CommonBuildDirectory.ENTRY_IDX_NAME]].append(
                f'{entry[CommonBuildDirectory.ENTRY_IDX_SNO]}:{entry[CommonBuildDirectory.ENTRY_IDX_ID]}'
            )
        self.__dir["resources"] = [sorted(buildEntryGIDsByName[resName]) for resName in sorted(buildEntryGIDsByName)]
        buildsWGT1EntryInStation = sum(1 for res in self.__dir["resources"] if len(set(b.split(":")[0] for b in res)) != len(res))
        if buildsWGT1EntryInStation:
            raise Exception(f"Can't rebuild as {buildsWGT1EntryInStation} builds have > 1 entry in some VistA")
        print(f'From {len(self.__dir["entries"]):,}, isolated {len(self.__dir["resources"]):,} distinct builds')
        
    # => won't keep descr in all etc. BUT still not too big!
    # ... want descr, date etc same OR note if > 1 form
    def _makeNormalizedEntries(self):
        pass
        
    def __flush(self):
        print("Builds {:,}, Entries {:,} flushed".format(len(self.__dir["resources"]), len(self.__dir["entries"])))
        json.dump(
            self.__dir, 
            open(self.__dirFile, "w"), 
            indent=4
        )
        
# ############################## DRIVER ##########################
                
def main():

    assert sys.version_info >= (3, 6)
    
    visn20VistAs = vistasOfVISNByOne("20")
    fileDirectory = CommonFileDirectory(visn20VistAs)
    cbd = CommonBuildDirectory(visn20VistAs, dirPrefixLetter="B")
    print(cbd.report(fileDirectory=fileDirectory))
            
if __name__ == "__main__":
    main() 


from setuptools import setup

setup(name='fmqlreports',
      description = 'FMQL Reports',
      long_description = """A Python framework and set of executables for generating specific reports on VistA data cached using FMQL and reduced using fmqlutils.""",
      version='0.7',
      install_requires = [
          'fmqlutils>=3.0',
          'python-dateutil'
      ],
      # install_requires=["pytz"], - problem OSX
      classifiers = ["Development Status :: 4 - Beta", "Programming Language :: Python :: 3"],
      url='http://github.com/Caregraf',
      license='Apache License, Version 2.0',
      keywords='VistA,FileMan,CHCS,FMQL,Caregraf',
      package_dir = {'fmqlreports': ''},
      packages = ['fmqlreports', 'fmqlreports.basics', 'fmqlreports.patient', 'fmqlreports.user', 'fmqlreports.imaging', 'fmqlreports.visit', 'fmqlreports.ifc', 'fmqlreports.directory'],
      entry_points = {
      }
)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime, date
from dateutil.relativedelta import relativedelta

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent, muBVC
from fmqlutils.cacher.cacherUtils import metaOfVistA, BASE_LOCN_TEMPL
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue, combineSubTypes, muBVCOfSTProp
from fmqlreports.webReportUtils import TOP_MD_TEMPL, SITE_DIR_TEMPL, keyStats, flattenFrequencyDistribution, roundFloat, reduce200
from .userClassifier import UserClassifier

"""
TODO:
- # of days (re-adjust reduction to see)
- order by duration x so where so at least count of 1 (even if duration 0)
- treat DoD special? ie/ 200_1?
- locals how? (and mixes?)
  - also issue of 'remote' but from local station?
- turn README into Excel Header (will do a combo excel, wiki instruction for import)
"""

"""
Subset of Report - CSV different options
"""
def csvUser(stationNo, deidentify=False):

    print("Preparing to make csv for {}{} - loading data ...".format(stationNo, " [DEIDENTIFY]" if deidentify else ""))
    
    if not os.path.isdir(BASE_LOCN_TEMPL.format(stationNo)):
        raise Exception("No setup for Station {}".format(stationNo))
    def ensureLocation(locn):
        if not os.path.isdir(locn):
            os.mkdir(locn)
        return locn
    csvDirBase = BASE_LOCN_TEMPL.format(stationNo) + "CSV/"
    ensureLocation(csvDirBase)
    csvDir = csvDirBase + "Users{}/".format("" if not deidentify else "DEID")
    ensureLocation(csvDir)
        
    meta = metaOfVistA(stationNo)
    if not ("name" in meta and "cutDate" in meta):
        raise Exception("Don't expect to be here if can't know name or cut date")
    vistaName = meta["name"]
    cutDate = meta["cutDate"]
    cutDateDT = datetime.strptime(cutDate.split("T")[0], "%Y-%m-%d")
    day90DateDT = cutDateDT - relativedelta(days=90)
    onAndAfterDay = datetime.strftime(day90DateDT, "%Y-%m-%d")
    upToDay = cutDate
    lastDayDT = cutDateDT - relativedelta(days=1)
    lastDay = datetime.strftime(lastDayDT, "%Y-%m-%d")
        
    dumpREADME(csvDir, vistaName, stationNo, cutDate, onAndAfterDay, lastDay, deidentify)

    allThere, details = checkDataPresent(stationNo, [
        # Additional settings/properties:
        # - duration (in seconds) 
        # - force count on 
        #     ipv4_address, remote_station_id, remote_user_ien, duration
        {"fileType": "3_081", "check": "DAY90E"},
        {"fileType": "200", "check": "ALL"} # Index
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    userInfoByUserRef = reduce200(stationNo) 
 
    """
    Consider: if need to break by device (ie/ segment off SSH etc? See sniff) and
    how to round out DOD USER breakdown ... are there any other 'CAPRI anyone' users
    that must be broken too ie/ not just one remote user => must break out for remote
    analysis ... 'is_remote_200_1'?
    """
    print("Loading subtypes ...")
    type3_081, st3_081ULRs = splitTypeDatas(stationNo, "3_081", reductionLabel="DAY90E", expectSubTypeProperties=["user", "level_of_assurance", "remote_app", "remote_200_user_ien", "workstation_label", "workstation_type", "remote_station_id", "device"])
    st3_081sByUserRef = defaultdict(list)
    for st in st3_081ULRs:
        userRef = singleValue(st, "user")
        st3_081sByUserRef[userRef].append(st)

    # Classify needs combo
    print("Recombining subtypes by user (backward compatible) ...")
    sts3_081Us = combineSubTypes(st3_081ULRs, ["user"], forceCountProps=["ipv4_address", "remote_station_id", "remote_user_ien", "duration"])
    print("... done")
    st3_081ByUserRef = dict((singleValue(st, "user"), st) for st in sts3_081Us if "user" in st)
    userClassifier = UserClassifier(stationNo, userInfoByUserRef, type3_081, st3_081ByUserRef)
    print("Classifying Users ...")
    classification = userClassifier.classify()
    print("... Classification complete")
        
    csvRemoteUsers(
        csvDir,
        classification["activeRemoteUserRefs"],
        userInfoByUserRef, 
        st3_081sByUserRef,
        st3_081ByUserRef,
        deidentify        
    )
    
def csvRemoteUsers(csvDir, activeRemoteUserRefs, userInfoByUserRef, st3_081sByUserRef, st3_081ByUserRef, deidentify):

    print("CSVing Remote Users ...")
    
    """
    Will move to classify
    """
    def extraEnforce(userRef, stsOfUser):
        for st in stsOfUser:
            if "remote_app" in st:
                remoteApp = singleValue(st, "remote_app")
                if re.match(r'MEDICAL DOMAIN WEB SERVICES', remoteApp) and singleValue(st, "level_of_assurance") != "1":
                    print(json.dumps(st, indent=4))
                    raise Exception("MDWS not LOA 1")
            elif singleValue(st, "level_of_assurance") != "1":
                print(json.dumps(st, indent=4))
                raise Exception("UNIDENTIFIED Remote App and LOA 2 - why remote?")
    
    remoteApps = sorted(list(set(
        singleValue(st, "remote_app") for userRef in st3_081sByUserRef for st in st3_081sByUserRef[userRef] if "remote_app" in st
    )))
    if sum(1 for userRef in st3_081sByUserRef for st in st3_081sByUserRef[userRef] if "remote_app" not in st):
        remoteApps.append("UNIDENTIFIED")
    nnRemoteApps = []
    for remoteApp in remoteApps:
        if re.match(r'MEDICAL DOMAIN WEB SERVICES', remoteApp):
            nnRemoteApp = "JLV"
        else:
            nnRemoteApp = re.sub(r'^VISTA IMAGING ', '', re.sub(r' DISPLAY$', '', remoteApp.split(" [")[0])).split("-")[0]
        nnRemoteApps.append(nnRemoteApp)
    
    cols = ["Last", "First", "SSN", "Local IEN", "Entered", "Remote Stations", "SOs", "First SO", "Last SO"]
    cols.extend(nnRemoteApps)
    
    fl = open("{}remoteUsers{}.csv".format(csvDir, "DEID" if deidentify else ""), "w")
    fl.write("{}\n".format(",".join([str(item) for item in cols])))
    
    print("Writing Remote Users ...")
    for i, userRef in enumerate(sorted(
        activeRemoteUserRefs, 
        key=lambda x: sum(st["_total"] for st in st3_081sByUserRef[x]),
        reverse=True
    )):
    
        extraEnforce(userRef, st3_081sByUserRef[userRef])
        
        userInfo = userInfoByUserRef[userRef] 
        userName = userRef.split(" [")[0]
        userNamePieces = userName.split(",")
        if len(userNamePieces) > 2:
            print("*** Only expected two pieces in names: {}".format(userName))
        lastName = userNamePieces[0]
        firstName = userNamePieces[1]
        if deidentify:
            firstName = firstName[0] + re.sub(r'[A-Za-z]', "X", firstName[1:])
            lastName = lastName[0] + re.sub(r'[A-Za-z]', "X", lastName[1:])
        userIEN = re.search(r'\-(\d+)\]', userRef).group(1)
        userEntered = userInfo["date_entered"] if "date_entered" in userInfo else ""
        userSSN = ("" if "ssn" not in userInfo else userInfo["ssn"]) if not deidentify else "XXXXXXXXX"
        row = [lastName, firstName, userSSN, userIEN, userEntered]

        remoteStationIds = sorted(list(set(remoteStationId if remoteStationId == "200CORP" else remoteStationId[0:3] for st in st3_081sByUserRef[userRef] for remoteStationId in st["remote_station_id"]["byValueCount"])))
        row.append("/".join(remoteStationIds))

        totalSignOns = sum(st["_total"] for st in st3_081sByUserRef[userRef])
        firstSignOn = sorted([st["date_time"]["firstCreateDate"].split("T")[0] for st in st3_081sByUserRef[userRef]])[0]
        lastSignOn = sorted([st["date_time"]["lastCreateDate"].split("T")[0] for st in st3_081sByUserRef[userRef]], reverse=True)[0]
        row.extend([totalSignOns, firstSignOn, lastSignOn])
        
        stByRemoteApp = dict((singleValue(st, "remote_app", ifMissingValue="UNIDENTIFIED"), st) for st in st3_081sByUserRef[userRef])
        for remoteApp in remoteApps:
            if remoteApp not in stByRemoteApp:
                row.append("")
                continue
            stRemoteApp = stByRemoteApp[remoteApp]
            if not ("duration" in stRemoteApp and "byValueCount" in stRemoteApp["duration"]):
                row.append(stRemoteApp["_total"])
                continue
            if stRemoteApp["_total"] > 1:
                kstatsDur = keyStats(
                    flattenFrequencyDistribution(stRemoteApp["duration"]["byValueCount"])
                )
                durMed = muSeconds(kstatsDur["median"])
            else:
                durMed = muSeconds(singleValue(stRemoteApp, "duration"))
            row.append("{} [{}]".format(stRemoteApp["_total"], durMed)) 
        
        fl.write("{}\n".format(",".join([str(item) for item in row])))
        
    print("... wrote {:,}".format(i))

    fl.close()
   
"""
TODO: may evolve to or do parallel CSV form so can put into cover page of an excel
""" 

README_TEMPL = """# Users from {} [{}]
{}
Data in Excel-compatible CSV about the users who logged into VistA Spokane over its latest 90 days. This VistA copy was cut on _{}_ so this user data is for the period of _{}_ through _{}_.

  * remoteUsers: users whose home VistA is elsewhere and who logged into this VistA using a range of applications from JLV to TeleReader during this 90 days. The first columns identify a user, with social security number, station number of the user's home VistA and the IEN of the user's record in this VistA's user file and the home VistA's user file. The application columns give the total number of signons for an application by a user and the median duration, day/hour/minute/second, is in square brackets. The reason JLV has such high numbers for signons and low durations is that a typical JLV session signs on 12 times into a VistA.
"""

README_DEID_QUAL = """
__Note__: "DEID" means de-identified which X's names and socials. An unredacted equivalent of this dataset is available.
"""

def dumpREADME(csvDir, vistaName, stationNo, cutDate, onAndAfterDay, lastDay, deidentify):

    README = README_TEMPL.format(
        vistaName,
        stationNo,
        
        README_DEID_QUAL if deidentify else "",
        
        cutDate,
        onAndAfterDay,
        lastDay
    )
    open(csvDir + "README_USERS{}.txt".format("_DEID" if deidentify else ""), "w").write(README)
    
# ######################### Utilities #######################
        
def muSeconds(seconds):
    seconds = int(seconds)
    (days, remainder) = divmod(seconds, 86400)
    (hours, remainder) = divmod(remainder, 3600)
    (minutes, seconds) = divmod(remainder, 60)
    if days:
        return "{} {}:{}:{}".format(
            int(days),
            int(hours),
            int(minutes),
            int(seconds)
        )
    if hours:
        return "{}:{}:{}".format(
            int(hours),
            int(minutes),
            int(seconds)
        )
    if minutes:
        return "{}:{}".format(
            int(minutes),
            int(seconds)
        )
    return int(seconds)
            
# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 6)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO [DEID]")

    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise SystemExit("Expect User Site to already exist with its basic contents")
    
    if len(sys.argv) == 3:
        deidentify = True
    else:
        deidentify = False
    
    csvUser(stationNo, deidentify)
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime
import platform
from random import randint

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.reporter.reportUtils import MarkdownTable, reportAbsAndPercent
from fmqlreports.webReportUtils import reduce200, vistasOfVISNByOne


"""
TODO FOR PATIENT/USER RE-ALIGN + PLOT EASY -- make the directories align and
preferably delegate/specialize a basic directory class ... just do the user dir
first and then do that.

LONGER GOAL: base class with PatientDirectory and simplify this so that like
Patient Directory has "usersFlat" and direct from 200 reduction etc.

TODO:
- md report
- for teler users, custom report on their overlaps (ie/ put into directory)
--------
- key is PROP cnt creator (most > 25+)/no creator (most <15) ... see breakdown as load and work out exceptions ... ie/ REMOTE vs LOCAL marker and more (termination goes with no creator?)
- sex for deid + deid for subset only ... make list and label mand
- more on later JLV (secid only, its context) vs earlier ie/ JLV stages
- isolate the HL7 user creation (any sig?)
- see TERMINATION ie/ how to gather "last sign in info" too ... possible to "purge" but must be clear ... align last sign in ... again push max 200 offers
- record date of clone in the vistas list
--------
- to fm.. inside/merge classify (then rename as 'makeTeleReaderUserDirectory' - see one function below to use the generic dir)

Observation from userReport.txt (from complete UserDirectory regen)
- have dateEntered for nearly all
- 46K of 324K explicit setup in 668 but 96K of 551K ie/ higher prop explicit in bigger 663 (ie/ less proportional remote setup which is to be expected)
- less ssn, more secid in 663, but seems to track explicit ie/ remotes still seem to be mainly ssn.
- other ids less important later: (but marker of unusual user?)
  - "distinct_unique_user_id" => not same as email or secid. High earlier (ie/ field used) but not many later (42/1K in 668; 79/1K for 663)
  - email only unique in very few of later entries (35/1K, 32/1K) 
- 648 only one that did bulk terminates (mainly of remotes? or?) and effects prop #'s and term props scew it up
"""
    
class CommonUserDirectory():

    ENTRY_SSN_IDX = 8
    ENTRY_SECID_IDX = 9
    ENTRY_NWNM_IDX = 10
    ENTRY_DUID_IDX = 11

    def __init__(self, vistaSNOs):
        self.__vistaSNOs = vistaSNOs
        self.__udirFile = f'{VISTA_DATA_BASE_DIR}Common/Directory/UDIR.json'
        try:
            print("Attempting to load preexisting User Directory")
            self.__udirInfo = json.load(open(self.__udirFile))
        except:
            print("Starting Directory from Scratch")
            self.__udirInfo = {"vistas": [], "entries": {}, "users": []}
        else:
            print("Loaded preexisting directory with {:,} users".format(len(self.__udirInfo["users"])))
        self.__changeUpdate()
        
    # ################## All Directory Accessors ######################
        
    def vistas(self):
        return self.__udirInfo["vistas"]
        
    # To see, per station, add stationNo
    def total(self, stationNo=""):
        if stationNo == "":
            return len(self.__udirInfo["users"])
        return sum(1 for user in self.__udirInfo["users"] if sum(1 for gien in user if gien.split(":")[0] == stationNo))
        
    def entryTotal(self, stationNo=""):
        if stationNo == "":
            return len(self.__udirInfo["entries"])
        return sum(1 for gien in self.__udirInfo["entries"] if gien.split(":")[0] == stationNo)
                    
    """
    Presence of SSN etc in one or more entries of users
    
    To see SSN
        noW, noWGT1 = .idPresence(CommonUserDirectory.ENTRY_SSN_IDX)
    same for ENTRY_SECID_IDX, ENTRY_NWNM_IDX, ENTRY_DUID_IDX
    """
    def idPresence(self, entryIdx, gt1=False):
        noW = 0
        noWGT1 = 0
        for i, user in enumerate(self.__udirInfo["users"], 1):
            vals = set(self.__udirInfo["entries"][gien][entryIdx] for gien in user if self.__udirInfo["entries"][gien][entryIdx])
            if len(vals):
                noW += 1
                if len(vals) > 1:
                    noWGT1 += 1
        return noWGT1 if gt1 else noW
        
    def gt1VistACount(self):
        cnt = 0
        for i, user in enumerate(self.__udirInfo["users"], 1):
            snos = set(gien.split(":")[0] for gien in user)         
            if len(snos) > 1:
                cnt += 1
        return cnt  
        
    def commonCounts(self, stationNo=""):
        """
        By count of vista's users are in, count types ex/
            4: 500 <---- in 4 VistAs
            3: 400 <---- in 3 VistAs
            ...
        
        Arguments:
        - stationNo: only do for files of one VistA

        For Plot: the common/plotCategoryBSV Plot and Explanation
        """
        commonCount = Counter()
        for user in self.__udirInfo["users"]:
            stationNos = set(gien.split(":")[0] for gien in user)
            if stationNo and stationNo not in stationNos:
                continue
            commonCount[len(stationNos)] += 1
        return commonCount # may not be in order ie/ 3, 1, 4, 2, ...
        
    """
    For a given id - ex SSN - return all values.

    Use: entryIDX = CommonUserDirectory.ENTRY_SSN_IDX ... ENTRY_SECID_IDX, 
    ENTRY_NWNM_IDX, ENTRY_DUID_IDX
    """
    def idValues(self, entryIdx):
        allVals = set()
        for i, user in enumerate(self.__udirInfo["users"], 1):
            vals = set(entries[gien][entryIdx] for gien in user if self.__udirInfo["entries"][gien][entryIdx])
            allVals |= vals
        return allVals
        
    # ###################### User Specific Methods #########################
        
    """
    TODO: promote to "All Directory"
    
    For time graphing
    
    Note: "ALL" => first date for a user irrespective
    """
    def entryDatesBySNO(self):
        dtBySNO = defaultdict(list)
        entries = self.__udirInfo["entries"]
        entryWODateBySNO = Counter()
        for gien in entries: # all entries now
            entry = entries[gien]
            sno = gien.split(":")[0]
            if entry[3] == "":
                entryWODateBySNO[sno] += 1
                continue
            dtBySNO[sno].append(entry[3])
        # Crude - ignoring order of orders - just resorting
        for sno in dtBySNO:
            dtBySNO[sno] = sorted(dtBySNO[sno])
        dtUsers = []
        for user in self.__udirInfo["users"]:
            entriesSDT = sorted([self.__udirInfo["entries"][gien][3] for gien in user if self.__udirInfo["entries"][gien][3]])
            if len(entriesSDT) == 0:
                entryWODateBySNO["USER"] += 1
                continue
            dtUsers.append(entriesSDT[0]) # take first date of the user
        dtBySNO["USER"] = dtUsers # overall uniques
        print("Serializing dates by station + USER overall:")
        for snop in dtBySNO:
            print("\t{}: {:,} - no date {:,}".format(snop, len(dtBySNO[snop]), entryWODateBySNO[snop]))
        return dtBySNO

    def lookupUserBySSN(self, ssn):
        try:
            self.__userBySSN
        except:
            self.__flattenUserData()
        if ssn in self.__userBySSN:
            userIndex = self.__userBySSN[ssn]
            userNames = self.__userNames[userIndex]
            # userNames one or many
            return userIndex, userNames
        return -1, []
        
    """
    This will fan out to entries that don't even have the SSN (or
    could have another!) but are pulled in by matching
    """
    def lookupUserEntriesBySSN(self, ssn):
        entriesOfSSN = []
        for user in self.__udirInfo["users"]:
            entriesOfUser = []
            gotOne = False
            for gien in user:
                entry = self.__udirInfo["entries"][gien]
                entriesOfUser.append(entry)
                if entry[CommonUserDirectory.ENTRY_SSN_IDX] == "":
                    continue
                if entry[CommonUserDirectory.ENTRY_SSN_IDX] == ssn:
                    gotOne = True
            if gotOne:
                entriesOfSSN.extend(entriesOfUser)
        return entriesOfSSN if len(entriesOfSSN) else None
        
    def userSSNs(self):
        return set(self.__udirInfo["entries"][gien][CommonUserDirectory.ENTRY_SSN_IDX] for gien in self.__udirInfo["entries"] if self.__udirInfo["entries"][gien][CommonUserDirectory.ENTRY_SSN_IDX])
        
    # Fan out the user - good for reports
    def lookupUserEntriesByGIEN(self, gien):
        gienUser = None
        for user in self.__udirInfo["users"]:
            if gien in user:
                gienUser = user
                break
        if not gienUser:
            return None
        entries = []
        for giene in gienUser:
            entry = self.__udirInfo["entries"][giene]
            entries.append(entry)
        return {"gien": gien, "entries": entries}
        
    # Can be > 1 as systems can have > 1
    def lookupGIENSByStationAndName(self, stationNo, name):
        giens = []
        for gien in self.__udirInfo["entries"]:
            if not re.match(stationNo, gien):
                continue
            entry = self.__udirInfo["entries"][gien]
            if entry[1] == name:
                giens.append(gien)
        return giens
        
    def lookupEntryByGIEN(self, gien):
        return self.__udirInfo["entries"].get(gien, None)
        
    """
    userGIENs => will fan out and if a user has any then it is pulled in as 
    are its other GIENs.
    
    TODO: move to work off usersFlat 
    
    TO ADD:
    - more on details (sex, termination etc ie/ PORTLAND more)
    - JLV and secid move ... ie/ SSN over time
    - date of VistA from first (.5) record + LAST RECORD date (as need for currency of report)
    """
    def report(self, userGIENs=None):
        countPerVistA = Counter()
        rec200CntPerSNO = Counter() # > count per VistA as dup 200's for some users 
        userRecordCount = Counter()
        cntGT1VistA = 0
        noWSSN = 0
        noWSECID = 0
        noWNWNM = 0
        noWDUID = 0
        noGT1NWNM = 0
        noGT1SSN = 0
        noGT1SECID = 0
        noGT1DUID = 0
        noNoSSNPlusNoSECIDPlusNoDUID = 0
        secids = set()
        nwnms = set()
        ssns = set()
        duids = set()
        perSNOHowMany = defaultdict(Counter) # per SNO, how many in other SNOs (graph)
        userIdIssuesByUserIndex = {}
        userNameIssueByUserIndex = {}
        if userGIENs:
            users = [user for user in self.__udirInfo["users"] if sum(1 for gien in user if gien in userGIENs)]
            inScopeGIENs = set(gien for user in users for gien in user)
            entries = dict((gien, self.__udirInfo["entries"][gien]) for gien in self.__udirInfo["entries"] if gien in inScopeGIENs)
        else:
            users = self.__udirInfo["users"]
            entries = self.__udirInfo["entries"]
        for i, user in enumerate(users, 1):
            userRecordCount[len(user)] += 1
            for gien in user:
                rec200CntPerSNO[gien.split(":")[0]] += 1
            snos = set(gien.split(":")[0] for gien in user)            
            for sno in snos:
                countPerVistA[sno] += 1
                perSNOHowMany[sno][len(snos)] += 1
            if len(snos) > 1:
                cntGT1VistA += 1
            tssns = set(entries[gien][CommonUserDirectory.ENTRY_SSN_IDX] for gien in user if entries[gien][CommonUserDirectory.ENTRY_SSN_IDX])
            tsecids = set(entries[gien][CommonUserDirectory.ENTRY_SECID_IDX] for gien in user if entries[gien][CommonUserDirectory.ENTRY_SECID_IDX])
            tnwnms = set(entries[gien][CommonUserDirectory.ENTRY_NWNM_IDX] for gien in user if entries[gien][CommonUserDirectory.ENTRY_NWNM_IDX])
            tduids = set(entries[gien][CommonUserDirectory.ENTRY_DUID_IDX] for gien in user if entries[gien][CommonUserDirectory.ENTRY_DUID_IDX])
            if len(tnwnms):
                noWNWNM += 1
                if len(tnwnms) > 1:
                    noGT1NWNM += 1
            if len(tssns):
                noWSSN += 1
                if len(tssns) > 1:
                    noGT1SSN += 1
            if len(tsecids):
                noWSECID += 1
                if len(tsecids) > 1:
                    noGT1SECID += 1
            if len(tduids):
                noWDUID += 1
                if len(tduids) > 1:
                    noGT1DUID += 1
            nwnms.update(tnwnms)
            ssns.update(tssns)
            secids.update(tsecids)
            duids.update(tduids)
            uentries = [entries[gien] for gien in user]
            idissues = []
            for idxd in [(CommonUserDirectory.ENTRY_SSN_IDX, "SSN"), (CommonUserDirectory.ENTRY_SECID_IDX, "SECID"), (CommonUserDirectory.ENTRY_DUID_IDX, "DUID")]:
                ids = set(entry[idxd[0]] for entry in uentries if entry[idxd[0]])        
                if len(ids) > 1:
                    namesNInitials = list(set("{} [{}]".format(entry[1], entry[2]) if entry[2] else entry[1] for entry in uentries))
                    idissues.append([idxd[1], list(ids), namesNInitials])
            if len(idissues):
                userIdIssuesByUserIndex[i] = idissues
            names = list(set(entry[1] for entry in uentries))
            if len(names) > 1: # Mainly one letter out or an initial
                userNameIssueByUserIndex[i] = names
            if len(tsecids) == 0 and len(tssns) == 0 and len(tduids) == 0:
                noNoSSNPlusNoSECIDPlusNoDUID += 1    
                    
        mu = ""
        mu += "# Cross-VistA User Directory\n\n"
        mu += "The following reports on a __Cross-VistA User Directory__ built from the User file (200) of <span class='yellowIt'>{:,}</span> VistAs, {}. Together they hold <span class='yellowIt'>{:,}</span> distinct users.\n\n".format(
            len(self.__vistaSNOs), # TODO: move into "vistas" (with POSTMASTER USER TOO)
            ", ".join(["{} [{}]".format(self.__vistaSNOs[sno], sno) for sno in sorted(self.__vistaSNOs, key=lambda x: int(x))]),
            len(users)
        )
        
        mu += """__Note__: this is the first version of this report. The presence of user records in more than one VistA, the dynamic addition of records to allow remote login and interfacility consults and the new move away from using SSNs and towards using VA's proprietary SECID as a user identifier, all allow for redundancy and contradiction across the VistAs. These mismatches and partial copying will be analyzed further in subsequent versions of this report.

"""
        
        tbl = MarkdownTable([":Property", "Value"], includeNo=False)
        tbl.addRow(["Users", "{:,}".format(len(users))])
        tbl.addRow(["Records", "{:,}".format(len(entries))])
        tbl.addRow(["In more than VistA", reportAbsAndPercent(cntGT1VistA, len(users))])
        tbl.addRow(["> 1 Record in an individual VistA", sum(1 for user in users if len(set(gien.split(":")[0] for gien in user)) != len(user))])
        tbl.addRow(["Has SSN", "{} [> 1 {:,}]".format(reportAbsAndPercent(noWSSN, len(users)), noGT1SSN)])
        tbl.addRow(["Has SECID", "{} [> 1 {:,}]".format(reportAbsAndPercent(noWSECID, len(users)), noGT1SECID)])
        tbl.addRow(["Has VA Network Name", "{} [>1 {:,}]".format(reportAbsAndPercent(noWNWNM, len(users)), noGT1NWNM)])
        tbl.addRow(["Has DUID", "{} [> 1 {}]".format(reportAbsAndPercent(noWDUID, len(users)), noGT1DUID)])
        tbl.addRow(["No SSN or SECID or DUID", reportAbsAndPercent(noNoSSNPlusNoSECIDPlusNoDUID, len(users))])
        mu += tbl.md() + "\n\n"
                                
        mu += "Per VistA ...\n\n"
        tbl = MarkdownTable([":VistA", "Users", "Records"], includeNo=False)
        for sno in sorted(countPerVistA, key=lambda x: countPerVistA[x], reverse=True):
            tbl.addRow([
                "__{} [{}]__".format(self.__vistaSNOs[sno], sno),
                reportAbsAndPercent(countPerVistA[sno], len(users)),
                reportAbsAndPercent(rec200CntPerSNO[sno], len(entries))
            ])  
        mu += tbl.md() + "\n\n"
        
        mu += "Across the {:,} VistAs, individual users have the following number of records. For example, a count for 3 records indicates how many users have 3 records ...\n\n".format(len(countPerVistA))
        tbl = MarkdownTable([":Records", "Users"], includeNo=False)
        for cnt in sorted(userRecordCount, key=lambda x: userRecordCount[x], reverse=True):
            tbl.addRow([
                "__{}__".format(cnt), 
                reportAbsAndPercent(userRecordCount[cnt], len(users))
            ])
        mu += tbl.md() + "\n\n"
        
        mu += "The following shows how many other VistAs also hold a user of a VistA. Note that dominant VistAs have proportionately higher single VistA counts - they alone hold certain users. Smaller VistAs contain users that also in other VistAs - they are rarely the only VistA with a user.\n\n"
        cols = [":VistA"]
        for i in range(1, len(countPerVistA) + 1):
            cols.append(str(i))
        tbl = MarkdownTable(cols, includeNo=False)
        for sno in sorted(perSNOHowMany, key=lambda x: sum(cnt for cnt in perSNOHowMany[x].values()), reverse=True):
            row = ["__{} [{}]__".format(self.__vistaSNOs[sno], sno)]
            for i in range(1, len(countPerVistA) + 1):
                if i not in perSNOHowMany[sno]:
                    row.append("")
                    continue
                row.append(reportAbsAndPercent(perSNOHowMany[sno][i], countPerVistA[sno]))
            tbl.addRow(row)
        mu += tbl.md() + "\n\n"
                
        return mu
        
    """
    Subset - used by TeleReader to produce sub directory (which it de-identifies)
    
        from deid import IDMaker, NameMaker
        sdeid = cud.subsetNDEID(allGIENs, IDMaker(), NameMaker())
    
    To see: pass in nm.report()
    
    preGeneratedNames = {"SNO:GIEN": FULL NAME - FIRST MID LAST} and then all
    other entries with that user ("SNO1:GIEN1" ...) will share this same name
    setup
    """
    def subsetNDEID(self, userGIENs, idMaker, nameMaker, preGeneratedNames={}):
        nusers = []
        nentries = {}
        snos = set()
        if len(preGeneratedNames):
            fnsExcluded = set(
               "{} {}".format(nm.split(" ")[0], nm.split(" ")[-1])
                for nm in preGeneratedNames.values()
            )
            nameMaker.excludeNames(fnsExcluded)            
        for i, user in enumerate(self.__udirInfo["users"]):
            if len(userGIENs) and sum(1 for gien in user if gien in userGIENs) == 0:
                continue
            nusers.append(user)
            # Only a Home User would have Sex
            sexes = set([self.__udirInfo["entries"][gien][4] for gien in user if self.__udirInfo["entries"][gien][4]])
            if len(sexes) > 1:
                raise Exception("Can't de id with SEX as > 1 Sex for a user")
            if len(sexes) == 0:
                sex = "female" if randint(1, 2) == 1 else "male" # fifty fifty
                print("** sexless user: {} - picking  {} 50/50".format(self.__udirInfo["entries"][user[0]][1], sex))
            else:
                sex = "male" if list(sexes)[0] == "M" else "female"
            nms = []
            if len(preGeneratedNames): # insist on full: first middle last
                for userGIEN in user:
                    if userGIEN in preGeneratedNames:
                        nms = preGeneratedNames[userGIEN].split(" ")
                        break
            if len(nms) == 0:
                nms = nameMaker.getNames(gender=sex, originalName=self.__udirInfo["entries"][user[0]][1])
            name, nameInitials, iname, inameInitials, nwnmEnd = nameMaker.calculateUserNameVariations(nms)
            origNWNM = ""
            for gien in user:
                snos.add(gien.split(":")[0])
                entry = self.__udirInfo["entries"][gien]
                nentries[gien] = entry
                nameToUse = iname if re.search(r' [A-Z]$', entry[1]) else name
                origName = entry[1]
                entry[1] = nameToUse
                if entry[2]:
                    entry[2] = inameInitials if nameToUse == iname else nameInitials
                for idxd in [(CommonUserDirectory.ENTRY_SSN_IDX, "SSN"), (CommonUserDirectory.ENTRY_SECID_IDX, "SECID"), (CommonUserDirectory.ENTRY_NWNM_IDX, "NWNM"), (CommonUserDirectory.ENTRY_DUID_IDX, "DUID")]:
                    if entry[idxd[0]] == "":
                        continue
                    if idxd[1] in ["SSN", "SECID"]:
                        entry[idxd[0]] = idMaker.nextID(idxd[1], entry[idxd[0]])
                        continue
                    if idxd[1] == "NWNM": 
                        # Part enforce, part repeat - saw same user with a BOI and 
                        # PUG VHA ID!
                        if origNWNM:
                            if entry[idxd[0]] != origNWNM:
                                print("** Warning > 1 NWNM - must be different starts: {} - {}".format(origNWNM, entry[idxd[0]]))
                            else:
                                origNWNM = entry[idxd[0]]
                        if re.match(r'VHA', entry[idxd[0]]): # allows for multi's in DEID too
                            nwnm = entry[idxd[0]][0:6] + nwnmEnd
                            entry[idxd[0]] = nwnm
                        else:
                            raise Exception("Only setup for VHA NW Names")     
                        continue
                    entry[idxd[0]] = "Y" # TODO
        print("Returning {:,} entries and {:,} users from User Directory{}".format(len(nentries), len(nusers), " [DEIDENTIFIED]" if nameMaker else ""))
        nvistas = [vinfo for vinfo in self.__udirInfo["vistas"] if vinfo["stationNumber"] in snos]
        return {"vistas": nvistas, "entries": nentries, "users": nusers}
        
    """
    QA Dump: nwnms
    
    TODO: more beyond VHA and get dups etc
    """
    def dumpNWNMs(self):
        
        vhaCounts = Counter()
        # '(VHA|VBA|VAF|VAHRC|VACO|VAH|VAMQA)' ... more than VHA and more than these too
        # ... but just do VHA for now
        # ... TODO: could [a] see NWNAME reuse which happens and [b] try tie to home sys
        for user in self.__udirInfo["users"]:
            nwnms = list(set(self.__udirInfo["entries"][gien][CommonUserDirectory.ENTRY_NWNM_IDX] for gien in user if self.__udirInfo["entries"][gien][CommonUserDirectory.ENTRY_NWNM_IDX]))
            vhaVistANWNames = set()
            for nwnm in nwnms:
                # nwname re.match('(VHA|VBA|VAF|VAHRC|VACO|VAH|VAMQA)', entry[idxd[0]]) <--- yep before do the events names
                nwnameMtch = re.match('VHA([A-Z]{3})', nwnm) 
                if nwnameMtch:
                    vhaVistANWNames.add(nwnameMtch.group(1))
            if len(vhaVistANWNames) == 1:
                vhaCounts[list(vhaVistANWNames)[0]] += 1
            elif len(vhaVistANWNames):
                vhaCounts["*VHAGT1*"] += 1
            elif len(nwnms):
                vhaCounts["*NON_VHAS_ONLY*"] += 1
            else:
                vhaCounts["*NONE*"] += 1
        print("NWNMs (VHA) [{:,}] - top 10:".format(len(vhaCounts)))
        for i, vha in enumerate(sorted(vhaCounts, key=lambda x: vhaCounts[x], reverse=True), 1):
            if i > 10:
                break
            print("\t{}: {:,}".format(vha, vhaCounts[vha]))
     
    """
    QA Dump: ID issues
    
    Overlaps, dups, inconsistencies across the Directory
    """
    def dumpIdIssues(self):
        userIdIssuesByUserIndex = {}
        userNameIssueByUserIndex = {}
        for i, user in enumerate(self.__udirInfo["users"], 1):
            entries = [self.__udirInfo["entries"][gien] for gien in user]
            idissues = []
            for idxd in [(CommonUserDirectory.ENTRY_SSN_IDX, "SSN"), (CommonUserDirectory.ENTRY_SECID_IDX, "SECID"), (CommonUserDirectory.ENTRY_DUID_IDX, "DUID")]:
                ids = set(entry[idxd[0]] for entry in entries if entry[idxd[0]])        
                if len(ids) > 1:
                    namesNInitials = list(set("{} [{}]".format(entry[1], entry[2]) if entry[2] else entry[1] for entry in entries))
                    idissues.append([idxd[1], list(ids), namesNInitials])
            if len(idissues):
                userIdIssuesByUserIndex[i] = idissues
            names = list(set(entry[1] for entry in entries))
            if len(names) > 1: # Mainly one letter out or an initial
                userNameIssueByUserIndex[i] = names
    
        print("Of {:,} users, {:,} have matching id issues (> 1 of same id), {:,} have matching name issues:".format(
            len(self.__udirInfo["users"]),
            len(userIdIssuesByUserIndex),
            len(userNameIssueByUserIndex)
        ))
        print("Id Issues:")
        for i in userIdIssuesByUserIndex:
            print("\t{}: {}".format(i, json.dumps(userIdIssuesByUserIndex[i])))
            
    """
    Index/flattening for lookups etc
    
    TODO: like patientDirectory - usersFlat
    """
    def __flattenUserData(self):
        self.__userBySSN = {}
        self.__userNames = []
        for i, user in enumerate(self.__udirInfo["users"]):
            tUserNames = set()
            for gien in user:
                entry = self.__udirInfo["entries"][gien]
                tUserNames.add(entry[1])
                if entry[CommonUserDirectory.ENTRY_SSN_IDX] == "":
                    continue
                essn = entry[CommonUserDirectory.ENTRY_SSN_IDX]
                self.__userBySSN[essn] = i      
            self.__userNames.append(sorted(list(tUserNames)))
            
    # ########################## CHANGE / UPDATE ######################
                        
    def __changeUpdate(self):
        changed = self.__fillFromAvailable()
        if changed:
            print("Changed - so rebuilding users")
            self.__rebuildUsers()
            self.__flattenUserData()
            self.__flush()    
                                
    def __fillFromAvailable(self):
        changed = False
        for stationNo in self.__vistaSNOs:
            if sum(1 for vistaInfo in self.__udirInfo["vistas"] if vistaInfo["stationNumber"] == stationNo):
                print("\t{} already in Directory - skipping".format(stationNo))
                continue
            try:
                usrs = reframeReportUsers(stationNo)
            except:
                print("Can't load User Reduction for {} as not available/creatable here".format(stationNo))
                raise
            changed = True
            print("Loaded USRs for {}: {:,}".format(stationNo, len(usrs)))
            print("Adding {} to directory".format(stationNo))
            # [gien, dateEntered, ssn, secid, distinctUniqueUserId] - in list, has index
            for info in usrs:
                entry = ["{}:{}".format(stationNo, info["ien"]), info.get("name", ""), info.get("initial", ""), info.get("dateEntered", ""), info["sex"].split(":")[0] if "sex" in info else "", "Y" if info.get("hasCreator") else "", "Y" if info.get("hasTerminationProp") else "", info.get("noProps"), info.get("ssn", ""), info.get("secid", ""), info.get("networkUserName", ""), info.get("distinctUniqueUserId", "")]
                self.__udirInfo["entries"][entry[0]] = entry
            self.__udirInfo["vistas"].append({"stationNumber": stationNo, "dateAdded": str(datetime.now()).split(" ")[0]})
        return changed
                
    """
    Any entries that share the same SSN or SECID (leaving DUID)

    TODO:
    - for deid, will replace actual ids with BOOLEAN of present or not in entries
    - add in true if CREATOR known
    - for n/w id, have an enum of type ie/ SPO, PUG etc. -- proxy for home
    - ... can save this as basis for report
    [NOTE: move to do DIRECTLY OFF USER RED and no intermediate]
    [DO FULL REPORT]
    [DO GRAPHIC ON IT]
    """
    def __rebuildUsers(self):
        del self.__udirInfo["users"]
        """
        Alg: establish users based on all ids of each entry, consistently
        - if id already allocated to use (by previous entry) then reuse that User
        - need two passes as if entry has > 1 id, different ones may have been
        allocated to separate users by previous entries that didn't share the same
        id combo (entry with only SSN, then one with only NWNM, then one with both)
        - second pass must see indexes that need to be combined

        Note: allows for many entries for a single user in a VistA (based on three
        key ids - SSN, SECID, NWNM).
                
        problem: SSN for User, then a SECID only user on own? and then both ...
        """
        userIndexesByGID = defaultdict(set)
        users = [[]] # only fixed entry for .5
        for gien in self.__udirInfo["entries"]:
            # Handle .5 manually
            if gien.split(":")[1] == ".5":
                users[0].append(gien)
                continue
            entry = self.__udirInfo["entries"][gien]
            # may have > 1 idx as diff ids may appear w/o each other in earlier entries
            userIndexes = set() 
            gids = set()
            
            # Can't enforce (5, "NWNM") as find duplicate use
            for idxd in [(CommonUserDirectory.ENTRY_SSN_IDX, "SSN"), (CommonUserDirectory.ENTRY_SECID_IDX, "SECID"), (CommonUserDirectory.ENTRY_DUID_IDX, "DUID")]:
                if entry[idxd[0]] == "":
                    continue
                gid = "{}:{}".format(idxd[1], entry[idxd[0]])
                gids.add(gid)
                if gid in userIndexesByGID:
                    userIndexes.add(userIndexesByGID[gid][0])                   
            if len(userIndexes) == 0:
                userIndexes = set([len(users)])
                users.append([])
            userIndex = sorted(list(userIndexes))[0] # pick first
            users[userIndex].append(gien)
            for gid in gids: 
                if gid not in userIndexesByGID:
                    userIndexesByGID[gid] = [userIndex] 
                    continue
                if userIndex not in userIndexesByGID[gid]:
                    userIndexesByGID[gid].append(userIndex)
        print("First Pass Users: {:,} users, {:,} gids and {:,} of the gids are in > 1 user".format(
            len(users),
            len(userIndexesByGID),
            sum(1 for gid in userIndexesByGID if len(userIndexesByGID[gid]) > 1)
        ))
        indexesToNix = set()
        cusers = []
        for gid in userIndexesByGID:
            if len(userIndexesByGID[gid]) == 1:
                continue
            cuser = set()
            for userIndex in userIndexesByGID[gid]:
                indexesToNix.add(userIndex)
                for gien in users[userIndex]:
                    cuser.add(gien)
            cuser = sorted(list(cuser))
            cusers.append(cuser)
        nusers = []
        for i, user in enumerate(users):
            if i in indexesToNix:
                continue
            nusers.append(user)
        nusers.extend(cusers)
        self.__udirInfo["users"] = nusers
        print("Second Pass Users: {:,} users".format(
            len(nusers)
        ))

    def __flush(self):
        print("Users {:,}, Entries {:,} flushed".format(len(self.__udirInfo["users"]), len(self.__udirInfo["entries"])))
        json.dump(
            self.__udirInfo, 
            open(self.__udirFile, "w"), 
            indent=4
        )
    
"""
TODO remove: Intermediate Form made to report/isolate better. Once know form
will nix this intermediate form.

ids: secid, ssn, [network_username] (tie to MUMPS/BSE - how created, how used)
"""
def reframeReportUsers(stationNo):

    userInfoByIEN = reduce200(stationNo)    
    
    print("Reframing/reporting user red {} - {:,}".format(stationNo, len(userInfoByIEN)))
    
    rinfos = []
    datesEntered = []
    noPropsHasCreator = Counter()
    noPropsNoCreator = Counter()
    for userRef in sorted(userInfoByIEN, key=lambda x: float(x.split("200-")[1].split("]")[0])):
        info = userInfoByIEN[userRef]
        ien = re.search(r'200\-([^\]]+)', userRef).group(1)
        # REM: reduce200 embeds name in id!
        rinfo = {"ien": ien, "name": userRef.split(" [")[0], "noProps": len(info["hasProps"])}
        if "initial" in info:
            rinfo["initial"] = info["initial"]
        if "date_entered" in info:
            rinfo["dateEntered"] = info["date_entered"]
            datesEntered.append(info["date_entered"])
        # Probably means "MACHINE" or "REMOTE"
        if "sex" in info:
            rinfo["sex"] = info["sex"]
        if "ssn" in info:
            rinfo["ssn"] = info["ssn"]
        if "secid" in info:
            rinfo["secid"] = info["secid"]
        if "network_username" in info:
            rinfo["networkUserName"] = info["network_username"]
        if "termination_date" in info["hasProps"] or "inactive_date" in info["hasProps"] or "termination_reason" in info["hasProps"]:
            rinfo["hasTerminationProp"] = True
        # if len(set(info["hasProps"]).intersection(set(["title", "creator", "primary_menu_option"]))) or len(info["hasProps"]) > 18: # may go 16 later
        #    rinfo["hasExplicitSetup"] = True
        if "creator" in info:
            rinfo["hasCreator"] = True
            noPropsToUse = "<25" if rinfo["noProps"] < 25 else "25+"
            noPropsHasCreator[noPropsToUse] += 1   
        else: # 20 for 648
            if "hasTerminationProp" not in rinfo:
                noPropsToUse = rinfo["noProps"] if rinfo["noProps"] < 15 else "15+"
            else:
                noPropsToUse = "HAS_TERMINATION_PROP"
            noPropsNoCreator[noPropsToUse] += 1                    
        if "__distinct_unique_user_id" in info:
            rinfo["distinctUniqueUserId"] = info["__distinct_unique_user_id"]
        if "is_subject_organization_vha" in info:
            rinfo["isSubjectOrganizationVHA"] = True
        rinfos.append(rinfo)
        
    print("__{}__: Returning {:,}; {} have date entered; {} have sex; {} has creator; {} has a \"termination property\"; {} have SSN ({:,} non 9 SSN, {:,} last 5000); {} network username ({}/last 5K); {} secid ({:,}/last 1K, {:,}/last 5K, {:,}/last 10K - org VHA {:,}); {} distinct unique user ids [special property normalized] ({:,} no secids, {:,}/last 1K, {:,}/last 5K)".format(
    
        stationNo,
        
        len(rinfos),
        reportAbsAndPercent(sum(1 for rinfo in rinfos if "dateEntered" in rinfo), len(rinfos)),
        reportAbsAndPercent(sum(1 for rinfo in rinfos if "sex" in rinfo), len(rinfos)),
        # Vast majority have 20+ props
        reportAbsAndPercent(sum(1 for rinfo in rinfos if "hasCreator" in rinfo), len(rinfos)),
        reportAbsAndPercent(sum(1 for rinfo in rinfos if "hasTerminationProp" in rinfo), len(rinfos)),
        
        reportAbsAndPercent(sum(1 for rinfo in rinfos if "ssn" in rinfo), len(rinfos)),
        sum(1 for rinfo in rinfos if "ssn" in rinfo and not re.match(r'\d{9}$', rinfo["ssn"])),
        sum(1 for rinfo in rinfos[-5000:] if "ssn" in rinfo),
        
        reportAbsAndPercent(sum(1 for rinfo in rinfos if "networkUserName" in rinfo), len(rinfos)),
        reportAbsAndPercent(sum(1 for rinfo in rinfos[-5000:] if "networkUserName" in rinfo), len(rinfos[-5000:])), 
        
        reportAbsAndPercent(sum(1 for rinfo in rinfos if "secid" in rinfo), len(rinfos)),
        sum(1 for rinfo in rinfos[-1000:] if "secid" in rinfo),
        sum(1 for rinfo in rinfos[-5000:] if "secid" in rinfo),
        sum(1 for rinfo in rinfos[-10000:] if "secid" in rinfo),
        sum(1 for rinfo in rinfos if "isSubjectOrganizationVHA" in rinfo),
        
        # DO MUMPS ISSUE
        reportAbsAndPercent(sum(1 for rinfo in rinfos if "distinctUniqueUserId" in rinfo), len(rinfos)),     
        sum(1 for rinfo in rinfos if "distinctUniqueUserId" in rinfo and "secid" not in rinfo),
        sum(1 for rinfo in rinfos[-1000:] if "distinctUniqueUserId" in rinfo and "secid" not in rinfo),
        sum(1 for rinfo in rinfos[-5000:] if "distinctUniqueUserId" in rinfo and "secid" not in rinfo)
    ))
    print("... Last date: {}".format(sorted(datesEntered)[-1]))
    print("... 5000 back has date {}".format(datesEntered[-5000]))
    print("# Props Has Creator:")
    for noProps in sorted(noPropsHasCreator, key=lambda x: noPropsHasCreator[x], reverse=True):
        print("\t{}: {}".format(noProps, noPropsHasCreator[noProps]))
    print("# Props No Creator:")
    for noProps in sorted(noPropsNoCreator, key=lambda x: noPropsNoCreator[x], reverse=True):
        print("\t{}: {}".format(noProps, noPropsNoCreator[noProps]))        
    return rinfos
    
# ############################## DRIVER ##########################
        
def main():

    assert sys.version_info >= (3, 6)
    
    cud = CommonUserDirectory(vistasOfVISNByOne("20"))
    print(cud.report())
            
if __name__ == "__main__":
    main()  
    


#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime
import platform

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.reporter.reportUtils import MarkdownTable, reportAbsAndPercent, muBVC
from fmqlreports.webReportUtils import reduce2, vistasOfVISNByOne

from userDirectory import CommonUserDirectory

"""
TODO:
- more on dumpIdIssues to tighten matches
- feed into matrix (by date) of yearly activity in key areas => current or not
  ... see: https://towardsdatascience.com/3-python-visualization-libraries-you-must-know-as-a-data-scientist-8d0cd25e1c73 at end on missingno and it showing nulls ... 
- feed back the UserDirectory -- may do common base class.
"""

class CommonPatientDirectory():

    ENTRY_FICN_IDX = 6
    ENTRY_ICN_IDX = 7
    ENTRY_SSN_IDX = 8
    ENTRY_PLID_IDX = 9
    ENTRY_SCN_IDX = 10
    
    IDINFOs = [
        (ENTRY_FICN_IDX, "FICN"),
        (ENTRY_ICN_IDX, "ICN"),
        (ENTRY_SSN_IDX, "SSN"),
        (ENTRY_PLID_IDX, "PLID"),
        (ENTRY_SCN_IDX, "SCN"),
    ]

    def __init__(self, vistaSNOs, userDirectory=None):
        self.__vistaSNOs = vistaSNOs
        self.__dirFile = f'{VISTA_DATA_BASE_DIR}Common/Directory/PDIR.json'
        try:
            print("Attempting to load preexisting Patient Directory")
            self.__pdirInfo = json.load(open(self.__dirFile))
        except:
            print("Starting Directory from Scratch")
            self.__pdirInfo = {"vistas": [], "entries": {}, "patients": []}
        else:
            print("Loaded preexisting directory with {:,} patients".format(len(self.__pdirInfo["patients"])))
        self.__changeUpdate(userDirectory)
        
    def report(self):
        mu = self.__report()
        return(mu)
        
    # ################## All Directory Accessors ######################
    # ... same as UserDirectory -- belongs in common base
        
    def vistas(self):
        return self.__pdirInfo["vistas"]
        
    # To see, per station, add stationNo
    def total(self, stationNo=""):
        if stationNo == "":
            return len(self.__pdirInfo["patients"])
        return sum(1 for patient in self.__pdirInfo["patients"] if sum(1 for gien in patient if gien.split(":")[0] == stationNo))
        
    def entryTotal(self, stationNo=""):
        if stationNo == "":
            return len(self.__pdirInfo["entries"])
        return sum(1 for gien in self.__pdirInfo["entries"] if gien.split(":")[0] == stationNo)
                    
    """
    Presence of SSN etc in one or more entries of patients
    
    To see SSN
        noW, noWGT1 = .idPresence(CommonUserDirectory.ENTRY_SSN_IDX)
    same for ENTRY_SECID_IDX, ENTRY_NWNM_IDX, ENTRY_DUID_IDX
    """
    def idPresence(self, entryIdx, gt1=False):
        noW = 0
        noWGT1 = 0
        for i, patient in enumerate(self.__pdirInfo["patients"], 1):
            vals = set(self.__pdirInfo["entries"][gien][entryIdx] for gien in patient if self.__pdirInfo["entries"][gien][entryIdx])
            if len(vals):
                noW += 1
                if len(vals) > 1:
                    noWGT1 += 1
        return noWGT1 if gt1 else noW
        
    """
    Number w/o any ID
    """
    def noIds(self):
        cnt = 0
        for i, patient in enumerate(self.__pdirInfo["patients"], 1):
            has = False
            for gien in patient:
                if sum(1 for idx in [
                    CommonPatientDirectory.ENTRY_FICN_IDX,
                    CommonPatientDirectory.ENTRY_ICN_IDX,
                    CommonPatientDirectory.ENTRY_SSN_IDX,
                    CommonPatientDirectory.ENTRY_PLID_IDX,
                    CommonPatientDirectory.ENTRY_SCN_IDX
                ] if self.__pdirInfo["entries"][gien][idx]):
                    has = True
                    break
            if has == False:
                cnt += 1
        return cnt
        
    def gt1VistACount(self):
        cnt = 0
        for i, patient in enumerate(self.__pdirInfo["patients"], 1):
            snos = set(gien.split(":")[0] for gien in patient)         
            if len(snos) > 1:
                cnt += 1
        return cnt  
        
    def commonCounts(self, stationNo=""):
        """
        By count of vista's patients are in, count types ex/
            4: 500 <---- in 4 VistAs
            3: 400 <---- in 3 VistAs
            ...
        
        Arguments:
        - stationNo: only do for files of one VistA

        For Plot: the common/plotCategoryBSV Plot and Explanation
        """
        commonCount = Counter()
        for patient in self.__pdirInfo["patients"]:
            stationNos = set(gien.split(":")[0] for gien in patient)
            if stationNo and stationNo not in stationNos:
                continue
            commonCount[len(stationNos)] += 1
        return commonCount # may not be in order ie/ 3, 1, 4, 2, ...
        
    """
    For a given id - ex SSN - return all values.

    Use: entryIDX = CommonPatientDirectory.ENTRY_FICN_IDX ... ENTRY_ICN_IDX, 
    ENTRY_SSN_IDX, ENTRY_PLID_IDX, ENTRY_SCN_IDX
    """
    def idValues(self, entryIdx):
        allVals = set()
        for i, patient in enumerate(self.__pdirInfo["patients"], 1):
            vals = set(entries[gien][entryIdx] for gien in patient if self.__pdirInfo["entries"][gien][entryIdx])
            allVals |= vals
        return allVals
        
    # ###################### Specifics #################
                        
    def deceased(self): # if any record marks individual as deceased then true
        cnt = 0
        for i, patient in enumerate(self.__pdirInfo["patients"], 1):
            has = False
            for gien in patient:
                if self.__pdirInfo["entries"][gien][5] == "Y":
                    has = True
                    break
            if has:
                cnt += 1
        return cnt
        
    def usersToo(self):
        if "sameAsUserBySSN" not in self.__pdirInfo:
            raise Exception("no sameUser calc for pdir so can't do Users Too")
        return len(self.__pdirInfo["sameAsUserBySSN"])
        
    def males(self): # all must be male
        cnt = 0
        for i, patient in enumerate(self.__pdirInfo["patients"], 1):
            sexes = set([self.__pdirInfo["entries"][gien][2] for gien in patient if self.__pdirInfo["entries"][gien][2]])
            if len(sexes) != 1:
                continue
            if list(sexes)[0] == "M":
                cnt += 1
        return cnt
        
    # ###################### Others (clean up) ####################
        
    """
    For time graphing
    
    TODO: may pad and guess for no dates that == between next ahead and last behind
    
    Big "NO DATE" problem ... 
    - 159K out of 960K Patients
    - 100K of 515K entries in PUG etc
    """
    def entryDatesBySNO(self):
        dtBySNO = defaultdict(list)
        entries = self.__pdirInfo["entries"]
        entryWODateBySNO = Counter()
        for gien in entries:
            entry = entries[gien]
            sno = gien.split(":")[0]
            if entry[3] == "": 
                entryWODateBySNO[sno] += 1
                continue
            dtBySNO[sno].append(entry[3])
        # Crude - ignoring order of orders - just resorting
        for sno in dtBySNO:
            dtBySNO[sno] = sorted(dtBySNO[sno])
        dtPatients = []
        for patient in self.__pdirInfo["patients"]:
            entriesSDT = sorted([self.__pdirInfo["entries"][gien][3] for gien in patient if self.__pdirInfo["entries"][gien][3]])
            if len(entriesSDT) == 0:
                entryWODateBySNO["PATIENT"] += 1
                continue
            dtPatients.append(entriesSDT[0]) # take first date of the patient
        dtBySNO["PATIENT"] = dtPatients # overall uniques
        print("Serializing dates by station + PATIENT overall:")
        for snop in dtBySNO:
            print("\t{}: {:,} - no date {:,}".format(snop, len(dtBySNO[snop]), entryWODateBySNO[snop]))
        return dtBySNO
        
    """
    Simple lookup by GIEN
    """
    def lookupEntryByGIEN(self, gien):
        self.__pdirInfo["entries"].get(gien, None)
        
    """
    Full (flat) lookup by GIEN
    """
    def lookupFPatientByGIEN(self, gien):
        try:
            self.__byGIEN
        except:
            patientsFlat = self.__flattenPatients()
            self.__byGIEN = {}
            for i, pf in enumerate(patientsFlat):
                for gien in pf[0]:
                    self.__byGIEN[gien] = (i, pf)
        if gien in self.__byGIEN:
            idx, pf = self.__byGIEN[gien]
            sameAsIDX = True if idx in self.__pdirInfo["sameAsUserBySSN"] else False
            pf.append(sameAsIDX)
            return pf
        return None
        
    """
    QA Dump: ID issues
    
    Total: 574
    W One Name ...: 280
    W Many Sexes ...: 85 <---- should break and ignore SSN match? TODO
    
    TODO:
    - more on multi SSNs
    - tie into deceased
    """
    def dumpIdIssues(self):
        mu = "## ID Issues - > 1 GICN\n"
        cntMFICNs = 0
        cntSameName = 0
        cntMultiSexes = 0
        patientsFlat = self.__flattenPatients()
        for pf in patientsFlat:
            if pf[CommonPatientDirectory.ENTRY_FICN_IDX] and len(pf[CommonPatientDirectory.ENTRY_FICN_IDX]) > 1:
                cntMFICNs += 1
                if len(pf[1]) == 1 or len(set(name.split(",")[0] for name in pf[1])) == 1:
                    cntSameName += 1
                if len(pf[2]) == 2:
                    cntMultiSexes += 1
        mu += "Total: {:,}\n".format(cntMFICNs)
        mu += "W One Name or Same Last Name (good indication that right): {:,}\n".format(cntSameName)
        mu += "W Many Sexes (strong mismatch?): {:,}\n".format(cntMultiSexes)
        print(mu)
        
    """
    Subset - used by TeleReader to produce sub directory (which it de-identifies)
    
        from deid import IDMaker, NameMaker
        sdeid = cpd.subsetNDEID(allGIENs, IDMaker(), NameMaker())
        
    Bonus: keeps ZZ if in original name AND leaves machine names (XX,PATIENT...) as is
    
    Handles (as here or DEID does)
    - middle initial or full middle name or no middle at all 
    - gender appropriate first and middle names
    - suffix (JR, II) of 2 part or three part names
    
    Doesn't handle (DEID doesn't or this doesn't)
    - > 3 elements (other than JR, SR, I, II ...)
    
    Roughly same as User Equivalent.
    
    TODO: could move to simpler, get name form from firstEntry as ALL ENTRIES seem
    to have same name (unlike User). Must re-inforce in Patient Report First.
    """
    def subsetNDEID(self, patientGIENs, idMaker, nameMaker, preGeneratedNames={}):
        if len(preGeneratedNames):
            fnsExcluded = set(
               "{} {}".format(nm.split(" ")[0], nm.split(" ")[-1])
                for nm in preGeneratedNames.values()
            )
            nameMaker.excludeNames(fnsExcluded)
        npatients = []
        nentries = {}
        snos = set()
        for i, patient in enumerate(self.__pdirInfo["patients"]):
            if len(patientGIENs) and sum(1 for gien in patient if gien in patientGIENs) == 0:
                continue
            # TODO: add to nameMaker to reuse names already given to same patient
            if i in self.__pdirInfo["sameAsUserBySSN"]:
                raise Exception("Haven't a/ced for Patient same as User in DEID name yet")
            npatients.append(patient)
            fentry = self.__pdirInfo["entries"][patient[0]]
            sexes = set([self.__pdirInfo["entries"][gien][2] for gien in patient if self.__pdirInfo["entries"][gien][2]])
            if len(sexes) > 1:
                # Some MEANS TEST, TEST PATIENTs are F in one, M in another
                if re.search(r'PATIENT', fentry[1]):
                    sexes = set(["M"])
                else:
                    raise Exception("Can't de id with SEX as > 1 Sex for a patient: {} - {}".format(json.dumps(patient), json.dumps(fentry)))
            if len(sexes) == 0:
                sex = "female" if randint(1, 5) == 1 else "male" # 20% female
                print("** sexless patient: {} - picking {} (Female 20%)".format(self.__pdirInfo["entries"][patient[0]][1], sex))
            else:
                sex = "male" if list(sexes)[0] == "M" else "female"
            keepOrigName = True if re.search(r'(PATIENT|TEST|\,VET)', fentry[1]) else False
            if not keepOrigName:   
                nms = []
                if len(preGeneratedNames): # insist on full: first middle last
                    for patientGIEN in patient:
                        if patientGIEN in preGeneratedNames:
                            nms = preGeneratedNames[patientGIEN].split(" ")
                            break
                if len(nms) == 0:
                    nms = nameMaker.getNames(gender=sex, originalName=fentry[1])
                sname, iname, fname = nameMaker.calculatePatientNameVariations(nms)
                namesUsed = set() # see if Patient, like User varies - doesn't seem to
            else:
                print("** Keeping Original Patient Name {}".format(fentry[1]))
            warned = False
            for gien in patient:
                snos.add(gien.split(":")[0])
                entry = self.__pdirInfo["entries"][gien]
                nentries[gien] = entry
                for idxd in CommonPatientDirectory.IDINFOs:
                    if entry[idxd[0]] == "":
                        continue
                    if idxd[1] in ["SSN", "FICN", "ICN"]:
                        entry[idxd[0]] = idMaker.nextID(idxd[1], entry[idxd[0]])
                        continue
                    # PLID, SCN as ignoring for now
                    entry[idxd[0]] = "Y" # TODO
                if keepOrigName:
                    continue
                origName = entry[1]
                namesUsed.add(origName)
                pieces = re.split(r',| ', origName)
                suffix = ""
                if pieces[-1] in ["SR", "JR", "I", "II", "III", "IV", "V"]:
                    suffix = pieces.pop()
                # May be crude but nixing last bit of "SMITH JAMES L K" etc.
                if len(pieces) > 3:
                    if not warned:
                        print("** Warning - DEID Patient Name {} doesn't account for its length - taking it as a three long name".format(origName))
                        warned = True # only warn once
                    while len(pieces) > 3:
                        pieces.pop()
                if len(pieces) == 3:
                    entry[1] = iname if len(pieces[2]) == 1 else fname
                else:
                    entry[1] = sname
                if suffix:
                    entry[1] = "{} {}".format(entry[1], suffix) 
                if re.match(r'ZZ', origName):
                    entry[1] = "ZZ{}".format(entry[1])
                    print("** Putting back ZZ on patient name: {} - {}".format(origName, entry[1]))
            # MAY USE THIS TO SIMPLIFY THE ABOVE ie/ firstEntry good enough for name calc
            if len(namesUsed) > 1:
                print("** Enforcing - though not relying on - that Patient has same name across entries - {}".format(json.dumps(list(namesUsed))))
        print("Returning {:,} entries and {:,} patients from Patient Directory{}".format(len(nentries), len(npatients), " [DEIDENTIFIED]" if nameMaker else ""))
        nvistas = [vinfo for vinfo in self.__pdirInfo["vistas"] if vinfo["stationNumber"] in snos]
        return {"vistas": nvistas, "entries": nentries, "patients": npatients}
        
    # REM: gien is a patient gien
    def sameAsUser(self, gien):
        patientsFlat = self.__flattenPatients()
        for idx in self.__pdirInfo["sameAsUserBySSN"]:
            pf = patientsFlat[int(idx)]
            if gien in pf[0]:
                return idx
        return ""
        
    # ########################## Report #####################
    
    """
    TODO: 
    - re-inforce in all entries for Patient have same name ie/ entry[1] the same.
    Seems to be true in deid above but see if generally so and if not, count the 
    exceptions (different from variation counter for users! ie/ much more Patient
    grooming/management)
    - isolate the synth and machine patients (ZZTEST ...?)
    """
    def __report(self, userGIENs=None):

        patientsFlat = self.__flattenPatients()
        entries = self.__pdirInfo["entries"].values()
                
        mu = ""
        mu += "# Cross-VistA Patient Directory\n\n"
        mu += "The following reports on a __Cross-VistA Patient Directory__ built from the Patient file (2) of <span class='yellowIt'>{:,}</span> VistAs, {}. Together they hold <span class='yellowIt'>{:,}</span> distinct patients.\n\n".format(
            len(self.__vistaSNOs), 
            ", ".join(["{} [{}]".format(self.__vistaSNOs[sno], sno) for sno in sorted(self.__vistaSNOs, key=lambda x: int(x))]),
            len(patientsFlat)
        )
        
        tbl = MarkdownTable([":Property", "Value"], includeNo=False)

        tbl.addRow(["Records", "{:,}".format(len(entries))])
        # Focus back on source records (may go in a final report)
        cntRecordBySNO = Counter()
        cntDTOutOfOrderBySNO = Counter()
        ienAndDTBySNO = defaultdict(list)  
        woDateBySNO = Counter() 
        wDeceasedBySNO = Counter()    
        for entry in entries:
            sno = entry[0].split(":")[0]
            cntRecordBySNO[sno] += 1
            if entry[5] == "Y":
                wDeceasedBySNO[sno] += 1
            if entry[3]:
                if len(ienAndDTBySNO[sno]) and entry[3] < ienAndDTBySNO[sno][-1]: 
                    cntDTOutOfOrderBySNO[sno] += 1
                ienAndDTBySNO[sno].append(entry[3])
            else:
                woDateBySNO[sno] += 1
        tbl.addRow([
            "Record per VistA", 
            muBVC(cntRecordBySNO)
        ])
        tbl.addRow([
            "Record with deceased indicated", 
            muBVC(wDeceasedBySNO)
        ])
        tbl.addRow([
            "Record w/o Date", 
            muBVC(woDateBySNO)
        ])
        tbl.addRow([
            "Record date out of order", 
            muBVC(cntDTOutOfOrderBySNO)
        ])
        tbl.addRow([
            "Record w/ names begin with ZZ (Deleted?)", 
            reportAbsAndPercent(
                sum(1 for entry in entries if re.match(r'ZZ', entry[1])),
                len(entries)
            )
        ])
         
        tbl.addRow(["Patients", "{:,}".format(len(patientsFlat))])
        tbl.addRow([
            "In more than VistA", 
            reportAbsAndPercent(
                sum(1 for pf in patientsFlat if len(set(e.split(":")[0] for e in pf[0])) > 1),
                len(patientsFlat)
            )
        ])
        tbl.addRow([
            "> 1 Record in an individual VistA", 
            reportAbsAndPercent(
                sum(1 for pf in patientsFlat if len(set(e.split(":")[0] for e in pf[0])) < len(pf[0])),
                len(patientsFlat)
            )
        ])
        tbl.addRow([
            "W/ Pseudo SSN", 
            reportAbsAndPercent(
                sum(1 for pf in patientsFlat if pf[CommonPatientDirectory.ENTRY_SSN_IDX] and sum(1 for e in pf[CommonPatientDirectory.ENTRY_SSN_IDX] if re.search(r'P$', e))),
                len(patientsFlat)
            ) 
        ])
        tbl.addRow([
            "Male", 
            reportAbsAndPercent( # not QA'ing > 1 val but should?
                sum(1 for pf in patientsFlat if pf[2] and pf[2][0] == "M"),
                len(patientsFlat)
            ) 
        ])
        tbl.addRow([
            "W/ deceased indicated somewhere", 
            reportAbsAndPercent(
                sum(1 for pf in patientsFlat if pf[5]),
                len(patientsFlat)
            ) 
        ])
        for idInfo in CommonPatientDirectory.IDINFOs:
            tbl.addRow([
                "Has {}".format(idInfo[1]),
                "{} [> 1 {:,}]".format(
                    reportAbsAndPercent(
                        sum(1 for pf in patientsFlat if pf[idInfo[0]]),
                        len(patientsFlat)
                    ),
                    sum(1 for pf in patientsFlat if pf[idInfo[0]] and len(pf[idInfo[0]]) > 1))
            ])
        tbl.addRow([
            "No ID", 
            reportAbsAndPercent(
                sum(1 for pf in patientsFlat if sum(1 for i in range(CommonPatientDirectory.ENTRY_FICN_IDX, len(pf)) if pf[i]) == 0),
                len(patientsFlat)
            )
        ])
        if "sameAsUserBySSN" in self.__pdirInfo:
            tbl.addRow(["Patients that are users ('sameAs')", reportAbsAndPercent(len(self.__pdirInfo["sameAsUserBySSN"]), len(patientsFlat))])
        mu += tbl.md() + "\n\n"
        
        patientCountPerVistA = Counter()
        recordCountPerVistA = Counter()
        recordCountPerPatientCount = Counter()
        perSNOHowMany = defaultdict(Counter)
        for pf in patientsFlat:
            recordCountPerPatientCount[len(pf[0])] += 1
            snos = set(gien.split(":")[0] for gien in pf[0])
            for sno in snos:
                patientCountPerVistA[sno] += 1
                recordCountPerVistA[sno] += len([gien for gien in pf[0] if re.match(sno, gien)])     
                perSNOHowMany[sno][len(snos)] += 1
        mu += "Per VistA ...\n\n"
        tbl = MarkdownTable([":VistA", "Patients", "Records"], includeNo=False)
        for sno in sorted(patientCountPerVistA, key=lambda x: patientCountPerVistA[x], reverse=True):
            tbl.addRow([
                "__{} [{}]__".format(self.__vistaSNOs[sno], sno),
                reportAbsAndPercent(patientCountPerVistA[sno], len(patientsFlat)),
                reportAbsAndPercent(recordCountPerVistA[sno], len(entries))
            ])  
        mu += tbl.md() + "\n\n"
        
        mu += "Across the {:,} VistAs, individual patients have the following number of records. For example, a count for 3 records indicates how many patients have 3 records ...\n\n".format(len(self.__pdirInfo["vistas"]))
        tbl = MarkdownTable([":Records", "Patient"], includeNo=False)
        for cnt in sorted(recordCountPerPatientCount, key=lambda x: recordCountPerPatientCount[x], reverse=True):
            tbl.addRow([
                "__{}__".format(cnt), 
                reportAbsAndPercent(recordCountPerPatientCount[cnt], len(patientsFlat))
            ])
        mu += tbl.md() + "\n\n"
        
        mu += "The following shows how many other VistAs also hold a patient of a VistA. Note that dominant VistAs have proportionately higher single VistA counts - they alone hold certain patients. Smaller VistAs contain patients that also in other VistAs - they are less likely to be the only VistA with a patient.\n\n"
        cols = [":VistA"]
        for i in range(1, len(patientCountPerVistA) + 1):
            cols.append(str(i))
        tbl = MarkdownTable(cols, includeNo=False)
        for sno in sorted(perSNOHowMany, key=lambda x: sum(cnt for cnt in perSNOHowMany[x].values()), reverse=True):
            row = ["__{} [{}]__".format(self.__vistaSNOs[sno], sno)]
            for i in range(1, len(patientCountPerVistA) + 1):
                if i not in perSNOHowMany[sno]:
                    row.append("")
                    continue
                row.append(reportAbsAndPercent(perSNOHowMany[sno][i], patientCountPerVistA[sno]))
            tbl.addRow(row)
        mu += tbl.md() + "\n\n"
        
        open("patientDirectoryReportN.md", "w").write(mu)
        
        return mu
        
    # ########################## Flatten Records ######################
    
    """
    Merges entries - replaces gien at start with plain sno
    """
    def __flattenPatients(self):
        if "patientsFlat" in self.__pdirInfo:
            return self.__pdirInfo["patientsFlat"]
        print("Flattening patients ...")
        self.__pdirInfo["patientsFlat"] = []            
        for patient in self.__pdirInfo["patients"]:
            for i, gien in enumerate(patient):
                entry = self.__pdirInfo["entries"][gien]
                if i == 0:
                    patientFlat = [[e] if e != "" else None for e in entry]
                    continue
                for j in range(0, len(patientFlat)):
                    if entry[j] == "":
                        continue
                    if patientFlat[j] == None:
                        patientFlat[j] = [entry[j]]
                        continue
                    if entry[j] in patientFlat[j]:
                        continue
                    patientFlat[j].append(entry[j])
                    patientFlat[j] = sorted(patientFlat[j])
            self.__pdirInfo["patientsFlat"].append(patientFlat)
        print("... finished")
        return self.__pdirInfo["patientsFlat"]
            
    # ########################## CHANGE / UPDATE ######################
            
    def __changeUpdate(self, userDirectory=None):
        changed = self.__fillFromAvailable()
        if changed:
            print("Changed - so rebuilding patients")
            self.__rebuildPatients()
            if "sameAsUserBySSN" in self.__pdirInfo: # set for update
                del self.__pdirInfo["sameAsUserBySSN"]    
        if userDirectory and "sameAsUserBySSN" not in self.__pdirInfo:
            self.__sameAsToUserDirectory(userDirectory)
            changed = True
        if changed:
            self.__flush()    
            self.dumpIdIssues()
                                
    def __fillFromAvailable(self):
        def reduceRed2Further(stationNo, red2):
            entry = []
            props = [
                ["gien", "{}:{}".format(stationNo, red2["ien"])], 
                "name", 
                ["sex", red2["sex"].split(":")[0] if "sex" in red2 else ""],
                "date_entered_into_file", 
                "preferred_facility", 
                ["dead", "Y" if ("date_of_death" in red2 or "death_entered_by" in red2) else ""], 
                "full_icn", 
                "integration_control_number", 
                "social_security_number", 
                "primary_long_id", 
                "subscription_control_number"
            ]
            for prop in props:
                if isinstance(prop, list):
                    entry.append(prop[1])
                    continue
                entry.append(red2.get(prop, ""))
            return entry
            
        changed = False
        for stationNo in self.__vistaSNOs:
            if sum(1 for vistaInfo in self.__pdirInfo["vistas"] if vistaInfo["stationNumber"] == stationNo):
                print("\t{} already in Directory - skipping".format(stationNo))
                continue
            try:
                patientInfoByIEN = reduce2(stationNo)    
            except:
                print("Can't load Patient Reduction for {} as not available/creatable here".format(stationNo))
                raise
            changed = True
            print("Loaded Patients for {}: {:,}".format(stationNo, len(patientInfoByIEN)))
            print("Adding {} to directory".format(stationNo))
            for ien in patientInfoByIEN:
                red2 = patientInfoByIEN[ien]
                entry = reduceRed2Further(stationNo, red2)
                self.__pdirInfo["entries"][entry[0]] = entry
            self.__pdirInfo["vistas"].append({"stationNumber": stationNo, "dateAdded": str(datetime.now()).split(" ")[0]})
        return changed
        
    """
    Any entries that share the same FICN, ICN (FICN missing) or SSN (non P)
    - for now, not using primary_long_id (only left if different from social with -'s)
    - primary_short_id == last 4 of social, again only for ref.
    
    Alg: establish patients based on all ids of each entry, consistently
    - if id already allocated to use (by previous entry) then reuse that Patient (WARN as unusual for Patient)
    - need two passes as if entry has > 1 id, different ones may have been
    allocated to separate users by previous entries that didn't share the same
        id combo (entry with only SSN, then one with only ICN, then one with both)
    - second pass must see indexes that need to be combined

    Note: allows for many entries for a single Patient in a VistA (based on two
    key ids - FICN, ICN (small part - few), SSN (non P)
    """
    def __rebuildPatients(self):
        del self.__pdirInfo["patients"]
        
        patientIndexesByGID = defaultdict(set)
        patients = [] # no fixed entries unlike user
        for gien in self.__pdirInfo["entries"]:
        
            entry = self.__pdirInfo["entries"][gien]

            # may have > 1 idx as diff ids may appear w/o each other in earlier entries
            patientIndexes = set() 
            gids = set()
            # No PLID or SCN for now
            for idxd in [(CommonPatientDirectory.ENTRY_FICN_IDX, "FICN"), (CommonPatientDirectory.ENTRY_ICN_IDX, "ICN"), (CommonPatientDirectory.ENTRY_SSN_IDX, "SSN")]:  
                if entry[idxd[0]] == "":
                    continue
                if idxd[1] == "SSN" and not re.match(r'\d{9}$', entry[idxd[0]]):
                    continue # P SSN or other invalid
                gid = "{}:{}".format(idxd[1], entry[idxd[0]])
                gids.add(gid)
                if gid in patientIndexesByGID:
                    patientIndexes.add(patientIndexesByGID[gid][0])                   
            if len(patientIndexes) == 0:
                patientIndexes = set([len(patients)])
                patients.append([])
            patientIndex = sorted(list(patientIndexes))[0] # pick first
            patients[patientIndex].append(gien)
            for gid in gids: 
                if gid not in patientIndexesByGID:
                    patientIndexesByGID[gid] = [patientIndex] 
                    continue
                if patientIndex not in patientIndexesByGID[gid]:
                    patientIndexesByGID[gid].append(patientIndex)
        print("First Pass Patients: {:,} patients, {:,} gids and {:,} of the gids are in > 1 patient".format(
            len(patients),
            len(patientIndexesByGID),
            sum(1 for gid in patientIndexesByGID if len(patientIndexesByGID[gid]) > 1)
        ))
        indexesToNix = set()
        cpatients = []
        for gid in patientIndexesByGID:
            if len(patientIndexesByGID[gid]) == 1:
                continue
            cpatient = set()
            for patientIndex in patientIndexesByGID[gid]:
                indexesToNix.add(patientIndex)
                for gien in patients[patientIndex]:
                    cpatient.add(gien)
            cpatient = sorted(list(cpatient))
            cpatients.append(cpatient)
        npatients = []
        for i, patient in enumerate(patients):
            if i in indexesToNix:
                continue
            npatients.append(patient)
        npatients.extend(cpatients)
        self.__pdirInfo["patients"] = npatients
        print("Second Pass Patients: {:,} patients".format(
            len(npatients)
        ))
        
    """
    UserDirectory has SSNs for many users. How many Patient entries are also users
    somewhere
    """
    def __sameAsToUserDirectory(self, userDirectory):
        def patientSSNs(patient):
            ssns = set()
            for gien in patient:
                entry = self.__pdirInfo["entries"][gien]
                if entry[CommonPatientDirectory.ENTRY_SSN_IDX]:
                    ssns.add(entry[CommonPatientDirectory.ENTRY_SSN_IDX])
            return ssns        
        self.__pdirInfo["sameAsUserBySSN"] = {} # reset
        for i, patient in enumerate(self.__pdirInfo["patients"]):
            ssns = patientSSNs(patient)
            for ssn in ssns:
                userIndex, userNames = userDirectory.lookupUserBySSN(ssn)
                if userIndex != -1:
                    self.__pdirInfo["sameAsUserBySSN"][i] = userIndex
        return self.__pdirInfo["sameAsUserBySSN"]

    def __flush(self):
        print("Patients {:,}, Entries {:,}, SameAs User {:,} flushed".format(len(self.__pdirInfo["patients"]), len(self.__pdirInfo["entries"]), len(self.__pdirInfo["sameAsUserBySSN"]) if "sameAsUserBySSN" in self.__pdirInfo else 0))
        json.dump(
            self.__pdirInfo, 
            open(self.__dirFile, "w"), 
            indent=4
        )
        
# ######################## Driver ###################
        
def main():

    assert sys.version_info >= (3, 6)
    
    visn20VistAs = vistasOfVISNByOne("20")
    cud = CommonUserDirectory(visn20VistAs)    
    cpd = CommonPatientDirectory(visn20VistAs, cud)
    print(cpd.report())
    json.dump(cpd.entryDatesBySNO(), open("Meta/Common/patientDatesBySNO.json", "w"), indent=4)
    # cpd.dumpSameAsUser()
            
if __name__ == "__main__":
    main()  
    